#ifdef __darwin__
#include <machine/endian.h>
#else
#include <endian.h>
#endif
#include <sys/stat.h>
#include <stdio.h>
#include <iostream>
#include <errno.h>
#include <Inventor/SbColor.h>
#include "rgbe.h"
#include "Main.h"
#include "FileUtils.h"
#include <vector>

#ifdef HAVE_OPENEXR
#include <ImfRgbaFile.h>
#endif
#include <png.h>

using namespace Imf;
using namespace std;

void missingEXRSupport(const char *file=0x0, const char *function=0x0){
  char blank[] = "";
  if (!file) file = blank;
  if (!function) function = blank;
  cerr << file << ":" << function << " "
       << "Missing OpenEXR support.\n"
       << "To enable EXR support, recompile with EXR=1 to make:\n"
       << "make EXR=1\n";
}

//'''''WW'''''''''''''WW''''WWWWW'''''''''''''WWW'''''''''''''''''''''''''''''
//     WW             WW    WW               WW  W                            
// WWW WWW   WW   WWW WW W  WWWW   WW   W WW WW  W WW W  WW  W WW WW    W W WW
//WW W WW W WWWW WW W WWW   WW    WW W  WW   WW  W WW W WWWW WW   WW WW W WW  
//WW   WW W WW   WW   WW W  WW    WW W  WW   WW  W WW W WW   WW   WW WW W WW  
// WWW WW W  WWW  WWW WW  W WW     WW   WW    WWW   WW   WWW WW    WW WW  WW  
//                                                                            
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

void FileUtils::checkForOverwrite(const string &filename,
                                  bool createBackup){

    // back up old file if it exists
  struct stat file_stat;
  if (stat(filename.c_str(), &file_stat)!=0 || !createBackup)
    return;

  // find unused filename
  string pre, ext, fn;
  string::size_type lastPeriod = filename.find_last_of('.');
  if (lastPeriod == string::npos)
    // insert number afterwards if no extension found
    pre = filename + ".";
  else {
    // insert number before extension
    pre = filename.substr(0,lastPeriod+1);
    ext = filename.substr(lastPeriod);
  }

  int i;
  for (i=1;;i++){

    char num[16];
    sprintf(num, "%03d", i);
    fn = pre + num + ext;
      
    if (stat(fn.c_str(), &file_stat))
      break;
  }

  // warn if many backups
  if (i>50)
    cerr << "Warning: excessive backups. You should probably delete some."
         << endl;

  // Hope there's no space in the filename!
  if (filename.find(' ') != string::npos){
    cerr << "Your filename has a space (skipping move cmd)\n";
    return;
  }

  // copy file
  string cmd = "mv " + filename + " " + fn;
  cerr << "Backing up old file..." << flush;
  cerr << cmd << endl;
  system(cmd.c_str());
}


//'''''''''''''WW''WW'''''''WW''W'WWWW''WWWW'
//                 WW       WW  W WW  W WW  W
//WW    W W WW WW WWWW  WW  WW  W WW  W WW  W
//WW WW W WW   WW  WW  WWWW WWWWW WW  W WWWW 
//WW WW W WW   WW  WW  WW   WW  W WW  W WW  W
// WW WW  WW   WW  WW   WWW WW  W WWWW  WW  W
//                                           
//'''''''''''''''''''''''''''''''''''''''''''

void FileUtils::writeHDR(const int imgDim[2], const string &imgFile, 
                         const SbColor *pix, bool createBackup){


  string fn;
  string::size_type lastPeriod = imgFile.find_last_of('.');
  if (lastPeriod == string::npos)
    fn = imgFile + ".hdr";
  else
    fn = imgFile.substr(0,lastPeriod) + ".hdr";

  checkForOverwrite(fn, createBackup);

  // write .hdr
  FILE *f = fopen(fn.c_str(), "wb");
  if (!f){
    cerr << "Error writing " << fn << ": " << strerror(errno) << endl;
    return;
  }
  RGBE_WriteHeader(f, imgDim[0], imgDim[1], NULL);
  for (int j=imgDim[1]-1; j>=0; j--)
    for (int i=0; i<imgDim[0]; i++){

      // .hdr's cannot have negative values
      const float *q = pix[j*imgDim[0]+i].getValue();
      float p[3];
      p[0] = q[0]<0 ? 0 : q[0];
      p[1] = q[1]<0 ? 0 : q[1];
      p[2] = q[2]<0 ? 0 : q[2];
      
      RGBE_WritePixels(f, p, 1);
    }
  fclose(f);
}

void FileUtils::writeRgba1(const string &filename,
                           const Rgba *pixels,
                           int width, int height){
#ifndef HAVE_OPENEXR
  missingEXRSupport(__FILE__, __FUNCTION__);
#else
 RgbaOutputFile file(filename.c_str(), width, height, WRITE_RGBA);
 file.setFrameBuffer(pixels, 1, width);
 file.writePixels(height);
#endif
}


void FileUtils::writeEXR(const int imgDim[2], const string &imgFile, 
                         const SbColor *pix, bool createBackup){
#ifndef HAVE_OPENEXR
  missingEXRSupport(__FILE__, __FUNCTION__);
#else
  string fn;
  string::size_type pos = imgFile.rfind('.');
  if (pos != string::npos){
    fn = imgFile.substr(0,pos) + ".exr";
  } else // no extension found
    fn = imgFile + ".exr";

  checkForOverwrite(fn, createBackup);

  vector<Rgba> rgb(imgDim[0]*imgDim[1]);
  for (int y=0; y<imgDim[1]; y++)
    for (int x=0; x<imgDim[0]; x++){
      int i=y*imgDim[0]+x;
      int j=(imgDim[1]-y)*imgDim[0]+x;
      rgb[j].r = pix[i][0];
      rgb[j].g = pix[i][1];
      rgb[j].b = pix[i][2];
      rgb[j].a = 1.;
    }
      
  // write .exr
  writeRgba1(fn, (const Rgba*)&rgb[0], imgDim[0], imgDim[1]);
#endif
}

bool FileUtils::writeDepth(const int imgDim[2], const string &imgFile,
                           const float *pixDepth, bool createBackup){

  string fn;
  string::size_type pos = imgFile.rfind('.');
  if (pos != string::npos){
    fn = imgFile.substr(0,pos) + "_depth.pfm";
  } else // no extension found
    fn = imgFile + "_depth.pfm";

  checkForOverwrite(fn.c_str(), createBackup);

#if __BYTE_ORDER == __LITTLE_ENDIAN
  const int byteOrder = -1;
#elif __BYTE_ORDER == __BIG_ENDIAN
  const int byteOrder = 1;
#endif
  
  FILE *f = fopen(fn.c_str(), "wb");
  if (!f){
    cerr << "Error writing " << fn << ": " << strerror(errno) << endl;
    return false;
  }

  fprintf(f, "Pf\n%d %d\n%d\n", imgDim[0], imgDim[1], byteOrder);
  
  for (int j=0; j<imgDim[1]; j++)
    for (int i=0; i<imgDim[0]; i++){
      float d = pixDepth[imgDim[0]*j+i];
      if (d<0) d = 1e50;
      fwrite(&d, sizeof(float), 1, f);
    }
  fclose(f);

  return true;
}


float* FileUtils::readPFM(const int imgDim[2], const string &filename,
                          bool verbose){

#if __BYTE_ORDER == __LITTLE_ENDIAN
  const int machineByteOrder = -1;
#elif __BYTE_ORDER == __BIG_ENDIAN
  const int machineByteOrder = 1;
#endif
  
  char line[4] = {0};
  size_t n;
  float scale;
  FILE *in = fopen(filename.c_str(), "rb");
  if (fscanf(in, "%2s\n", line)!=1){
    printf("Error reading pfm line 1.\n");
    return NULL;
  }
  if (strcasecmp("pf", line)){
    printf("Unrecognized file format.\n");
    printf("line=[%s]\n", line);
    return NULL;
  }
  if (fscanf(in, "%d %d\n", &imgDim[0], &imgDim[1])!=2){
    printf("Error reading pfm line 2.\n");
    return NULL;
  }
  if (fscanf(in, "%f\n", &scale)!=1){
    printf("Error reading pfm line 3.\n");
    return NULL;
  }

  bool revOrder = false;
  int fileByteOrder = scale<0 ? -1 : 1;
  if ((scale<0 && machineByteOrder>0) ||
      (scale>0 && machineByteOrder<0)){
    revOrder = true;
  }
  if (scale<0) scale = -scale;

  if (verbose){
    printf("Image is %dx%d with scale %g. ", imgDim[0], imgDim[1], scale);
    printf("Machine is %s, file is %s.\n",
           machineByteOrder == -1 ? "little endian" : "big endian",
           fileByteOrder == -1 ? "little endian" : "big endian");
  }
  
  float *depthRGB = new float[imgDim[0]*imgDim[1]*3];
  float depth;
  for (int i=0; i<imgDim[0]*imgDim[1]; i++){
    fread(&depth, sizeof(float), 1, in);
    if (revOrder){
      char *c = (char *)&depth;
      char o[4];
      o[3] = c[0];
      o[2] = c[1];
      o[1] = c[2];
      o[0] = c[3];
      depth = *(float*)o;
    }
    depth *= scale;
    depthRGB[i*3+0] = depth;
    depthRGB[i*3+1] = depth;
    depthRGB[i*3+2] = depth;
  }
  fclose(in);

  return depthRGB;
}


  // Old PPM and PFM code
//   // open file
//   out.open(gv->imgFile);
//   if (!out){
//     cerr << "Error writing file " << gv->imgFile << endl;
//     return;
//   }
//  
//   // print PPM identifier
//   if (gv->floatMap)
//     out << "Pf\n";
//   else
//     out << "P6\n";
//   out << comment.str() << "\n";
//   out << gv->imgDim[X] << " " << gv->imgDim[Y] << "\n";
//   if (!gv->floatMap)
//     out << 255 << "\n";
//  
//   // english reading order output
//   SbColor *c;
//   double r, g, b;
//   for (int j=gv->imgDim[Y]-1; j>=0; j--){
//     for (int i=0; i<gv->imgDim[X]; i++){
//       c = &pix[gv->imgDim[X]*j+i];
//       r = (*c)[R];
//       g = (*c)[G];
//       b = (*c)[B];
//
//       if (gv->floatMap)
// 	out << r << " "
// 	    << g << " "
// 	    << b << "\n";
//       else {
// 	unsigned char rc, gc, bc;
// 	rc = (unsigned char) clamp((int)(r*255.0), 0, 255);
// 	gc = (unsigned char) clamp((int)(g*255.0), 0, 255);
// 	bc = (unsigned char) clamp((int)(b*255.0), 0, 255);
//
// 	out.put(rc);
// 	out.put(gc);
// 	out.put(bc);
//       }
//      
//     }
//   }
//  
//   out.close();
//   cerr << "done." << endl;


bool FileUtils::writePNG(const int imgDim[], const string &filename,
                         const SbColor *pix, float gamma,
                         const string &comment){
                         
  // write PNG
  // source from http://www.libpng.org/pub/png/libpng-1.2.5-manual.html#section-4

  // open file
  FILE *fp = fopen(filename.c_str(), "wb");
  if (!fp){
    cerr << "Error writing file " << filename << endl;
    return false;
  }

  // create png_struct and png_info
  png_structp png_ptr = png_create_write_struct
    (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr)
    return false;

  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr){
    png_destroy_write_struct(&png_ptr,
			     (png_infopp)NULL);
    cerr << "Error creating PNG data structures." << endl;
    return false;
  }

  // set up error handling. PNG will expect to longjmp() here
  // not supported on solaris
#ifndef SUNOS
  if (setjmp(png_jmpbuf(png_ptr))){
       png_destroy_write_struct(&png_ptr, &info_ptr);
       fclose(fp);
    cerr << "PNG creation error." << endl;
    return false;
  }
#endif

  // set up png output
  png_init_io(png_ptr, fp);
  png_set_IHDR(png_ptr, info_ptr, imgDim[0],
	       imgDim[1], 8, PNG_COLOR_TYPE_RGB,
	       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
	       PNG_FILTER_TYPE_DEFAULT);

  // convert comment to c style string
  if (!comment.empty()){
    char *copy = strdup(comment.c_str());
  
    // create comment
    png_text text_ptr[1];
    text_ptr[0].key = "Comment";
    text_ptr[0].text = copy;
    text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
    png_set_text(png_ptr, info_ptr, text_ptr, 1);
    free(copy);
  }

  // set rendering intent
  //png_set_sRGB(png_ptr, info_ptr, PNG_sRGB_INTENT_ABSOLUTE);
  
  // set gamma
  png_set_gAMA(png_ptr, info_ptr, 1.0f/gamma);

  // set chromacity info (values from srgb.com)
  png_set_cHRM(png_ptr, info_ptr,
	       .3127, .3290,        // white point x, y
	       .64, .33,            // r 
	       .15, .06,            // g
	       .3127, .3290);       // b

  // create array of bytes
  png_byte *image = new png_byte[imgDim[0]*imgDim[0]*3];
  png_bytep row_pointers[imgDim[1]];
  const SbColor *c;
  double r, g, b;
  for (int j=0; j<imgDim[1]; j++){
    for (int i=0; i<imgDim[0]; i++){
      c = &pix[imgDim[0]*j+i];
      
      r = (*c)[0];
      g = (*c)[1];
      b = (*c)[2];
      
      unsigned char rc, gc, bc;
      rc = (unsigned char) clamp((int)(r*255.0), 0, 255);
      gc = (unsigned char) clamp((int)(g*255.0), 0, 255);
      bc = (unsigned char) clamp((int)(b*255.0), 0, 255);

      image[j*(imgDim[0]*3)+i*3+0] = rc;
      image[j*(imgDim[0]*3)+i*3+1] = gc;
      image[j*(imgDim[0]*3)+i*3+2] = bc;
    }
  }
  for (int k=0; k<imgDim[1]; k++)
    row_pointers[k] = &image[(imgDim[1]-k-1)*(imgDim[0]*3)];

  // write image
  png_write_info(png_ptr, info_ptr);
  png_write_image(png_ptr, row_pointers);
  png_write_end(png_ptr, info_ptr);

  // clean up
  png_destroy_write_struct(&png_ptr, &info_ptr);
  fclose(fp);
  delete [] image;
  return true;
}
