////////////////////////////////////////////////////////////////////////////
//
//WWWW''''''''''''''''''WW'WW'''''''''''''''''''''''''WWW''''''''''''WW'''''''
// WW                   WW                           WW  W           WW       
// WW  W WW W WW  WWW  WWW WW  WWW WWW   WWW  WW     WW     WWW  WWW WWW   WW 
// WW  WW   WW   W WW W WW WW W WW WW W WW W WWWW    WW    W WW WW W WW W WWWW
// WW  WW   WW   W WW W WW WW W WW WW W WW   WW      WW  W W WW WW   WW W WW  
//WWWW WW   WW    WWW  WWW WW  WWW WW W  WWW  WWW     WWW   WWW  WWW WW W  WWW
//                                                                            
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// File: IrradianceCache.cxx
// Description: IrradianceCache class implementation
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2005 Kevin Beason
// 
// Irradiance Cache for the Pane ray tracer
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// This class is an implementation of an Irradiance Cache,
// which is described in the following paper:
//
// A Ray Tracing Solution for Diffuse Interreflection
// by Greg Ward.
//
// Briefly, irradiance samples are stored along
// with thier position in an octtree in such a way as to quickly
// find nearby stored samples for a given point in space.
//
// On querying, nearby samples are found (if they exist) and their
// values are interpolated to produce values between samples.
//
// The samples are interpolated using the gradient of the irradiance,
// as described in:
//
// Irradiance Gradients
// by Gregory J.. Ward and Paul Heckbert
// 
// $Id: IrradianceCache.cxx 1186 2007-11-12 09:02:06Z beason $
//
////////////////////////////////////////////////////////////////////////////

#include "IrradianceCache.h"
#include "Main.h"
#include "Ray.h"
#include "debug.h"
#include "GlobalVars.h"

//          ____
//        /|2  /|6    y
//       / |  / |     |
//      /__|0/  |4    |__x
//     3|  / |7 /    /
//      | /  | /    z
//     1|/___|5
//
// vertex list
const int IrradianceCache::delta[8][3] = {
  0, 0, 0,
  0, 0, 1,
  0, 1, 0,
  0, 1, 1,
  
  1, 0, 0,
  1, 0, 1,
  1, 1, 0,
  1, 1, 1
};


//'WWW''''''''''''''''''''WW''''''''''''''''''WW''''''''''''
//WW  W                   WW                  WW            
//WW     WW   WWW   WWWW WWWW W WW W WW  WWW WWWW  WW   W WW
//WW    WW W  WW W WWW    WW  WW   W WW WW W  WW  WW W  WW  
//WW  W WW W  WW W   WWW  WW  WW   W WW WW    WW  WW W  WW  
// WWW   WW   WW W WWWW   WW  WW    WWW  WWW  WW   WW   WW  
//                                                          
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Constructor
//
IrradianceCache::IrradianceCache(){
  min = SbVec3f(0.0, 0.0, 0.0);
  first_cell_size = 1.0;
  root = new ic_node;
  for (int i=0; i<8; i++)
    root->child[i] = NULL;
  root->terminal = 1;
  root->half_size = first_cell_size/2.0;
  root->mid = min + SbVec3f(first_cell_size/2.0,
			    first_cell_size/2.0,
			    first_cell_size/2.0);
  num_pnts = 0;
  treeDebug = 0;
  a = 1.0;
  resolution = 64;
  gv = NULL;
}


//'''''''WW''''''''''''''''''WWWW''''''''''''''''''WW'WW''''''''''''''''''''
//       WW                   WW                   WW                       
// WWWW WWWW  WW   W WW  WW   WW  W WW W WW  WWW  WWW WW  WWW WWW   WWW  WW 
//WWW    WW  WW W  WW   WWWW  WW  WW   WW   W WW W WW WW W WW WW W WW W WWWW
//  WWW  WW  WW W  WW   WW    WW  WW   WW   W WW W WW WW W WW WW W WW   WW  
//WWWW   WW   WW   WW    WWW WWWW WW   WW    WWW  WWW WW  WWW WW W  WWW  WWW
//                                                                          
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// stores irradiance value in cache
//
void IrradianceCache::storeIrradiance(SbVec3f &e, SbVec3f &p, SbVec3f &n,
				      SbVec3f &ng, float r, SbVec3f &Er,
				      SbVec3f &Et){

  const float minri = minRi;
  const float maxri = 16.0*minri;

  const float brightness = 0.27*e[R] + 0.67*e[G] + 0.06*e[B];
  float d = brightness ? Et.dot(Et)/(brightness*brightness) : 0.0;

  // cap r
  //float delt = sqrt(d*r*r);
  if (r>maxri)
    r = maxri;

  // reduce radius if gradient large
  if (d*r*r>1.0){
    r=1.0/sqrt(d);
  }


  // NOTE:
  //
  // the tests below are heavily imitating those
  // found in the Radiance source code:
  // http://radsite.lbl.gov/radiance/
  


//   // experimental rotatial gradient limitation
//   float md = 1.0-a*a;
//   md *= md;
//   md = sqrt(1.0-md);
//   float gd = sqrt(Er.dot(Er))/brightness;
//   if (gd*md > 0.5){
//     r = 1.0/gd;
//     showPoint(p, 0.01, SbColor(1,1,0));
//     cerr << " FOUND ONE \n";
//   }

  
  // set minimum r
  if (r<minri){
    r=minri;

    // cap gradient
    if (d*r*r>1.0){
      d=1.0/r/sqrt(d);
      Et*=d;
    }
  }

  if (DEBUG_STATE)
    cerr << "new r: " << r << "\n";
  
  float vd = a*r;

  // find or create cell with:   2*vd < size < 4*vd
  SbVec3f currmin = min, currmid;
  float currsize = first_cell_size;
  ic_node *currnode = root;
  char child;

  while (currsize > 4.0*vd){ // subdivide
    currsize /= 2.0;
    currmid = currmin + SbVec3f(currsize, currsize, currsize);
    child = 0;
    
    // X
    if (p[X] > currmid[X]){
      child |= 4;
      currmin[X] = currmid[X];
    }

    // Y
    if (p[Y] > currmid[Y]){
      child |= 2;
      currmin[Y] = currmid[Y];
    }

    // Z
    if (p[Z] > currmid[Z]){
      child |= 1;
      currmin[Z] = currmid[Z];
    }

    // create new child
    if (currnode->child[child] == NULL){
      currnode->terminal = 0;
      ic_node *newnode = new ic_node; // FIXME-SPEED
      for (int i=0; i<8; i++)
	newnode->child[i] = NULL;
      newnode->terminal = 1;
      newnode->half_size = currsize/2.0;
      newnode->mid = currmin + SbVec3f(currsize/2.0,
				       currsize/2.0,
				       currsize/2.0);
      currnode->child[child] = newnode;
      currnode = newnode;
      
      if (DEBUG_STATE)
	showCube(currmin, currmin+SbVec3f(currsize, currsize, currsize),
		 SbVec3f(0.5, 0.1, 0.1));
    }
    else // go to child
      currnode = currnode->child[child];
  }

  // FIXME: add test using procedure similar to findPoints
  // to detect if this point might be invalid.
  
  // store irradiance in current node
  irpoint *newpnt = new irpoint; // FIXME-SPEED
  newpnt->irrad = e;
  newpnt->ri = r;
  newpnt->p = p;
  newpnt->n = n;
  newpnt->ng = ng;
  newpnt->Er = Er;
  newpnt->Et = Et;

  // visualize
//   SbVec3f ern = Er; ern.normalize();
//   SbVec3f etn = Et; etn.normalize();
//   Ray er(p, ern); er.show2(0.1, SbColor(0.0, 1.0, 0.0));
//   Ray et(p, etn); et.show2(0.1, SbColor(1.0, 0.0, 0.0));


//  if (DEBUG_STATE)
//    cerr << "\nstoring irrad point in node " << currnode << "\n";

  currnode->pntlist.push_back(newpnt);
  
  num_pnts++;
}



//'''''''''''WW''WWWW''''''''''''''''''WW'WW''''''''''''''''''''
//           WW   WW                   WW                       
// WWW  WW  WWWW  WW  W WW W WW  WWW  WWW WW  WWW WWW   WWW  WW 
//W WW WWWW  WW   WW  WW   WW   W WW W WW WW W WW WW W WW W WWWW
//W WW WW    WW   WW  WW   WW   W WW W WW WW W WW WW W WW   WW  
// WWW  WWW  WW  WWWW WW   WW    WWW  WWW WW  WWW WW W  WWW  WWW
//  WW                                                          
//WWW'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// returns interpolated result if nearby cached values exist
//
int IrradianceCache::getIrradiance(SbVec3f *e, Ray &r, SbVec3f *n,
				   SbVec3f *ng, int prepass){
  
  // construct irr_calc to pass
  cacheParam icp;
  icp.e = e;
  icp.p = &r.x;
  icp.ctx = r.ctx;
  icp.n = n;
  icp.ng = ng;
  *e = SbVec3f(0.0, 0.0, 0.0);
  icp.denom = 0.0;
  icp.num_vals = 0;
  icp.exactFound = 0;
  icp.prepass = prepass;

  // traverse tree
  findPoints(icp, root);

  if (icp.num_vals){
    *e /= icp.denom;

    // sometimes the gradient makes the irradiance negative
    if ((*e)[0] < 0.0)
      (*e)[0] = 0.0;
    if ((*e)[1] < 0.0)
      (*e)[1] = 0.0;
    if ((*e)[2] < 0.0)
      (*e)[2] = 0.0;
    
    return(1);
  }
  else
    return(0);

};



//''WW'WW''''''''WW'WWWW''''''''WW'''''''WW'''''''
// WW            WW WW  W                WW       
//WWWW WW WWW   WWW WW  W  WW   WW WWW  WWWW  WWWW
// WW  WW WW W W WW WWWW  WW W  WW WW W  WW  WWW  
// WW  WW WW W W WW WW    WW W  WW WW W  WW    WWW
// WW  WW WW W  WWW WW     WW   WW WW W  WW  WWWW 
//                                                
//''''''''''''''''''''''''''''''''''''''''''''''''
//
// traverse octtree looking for points
//
void IrradianceCache::findPoints(cacheParam &icp, ic_node *currnode){

    if (currnode == NULL)
      return;

    if (0 && DEBUG_STATE)
      cerr << "searching node " << currnode << "\n";

    // loop over points at this node
    list<irpoint *>::iterator I = currnode->pntlist.begin();
    list<irpoint *>::iterator end = currnode->pntlist.end();
    
    for (; I != end; I++){
      irpoint *ip = *I;

      if (0 &&DEBUG_STATE){
	cerr << "examining point: "; printVec(ip->p); cerr << "\n";
      }

      // from
      // A Ray Tracing Solution for Diffuse Interreflection
      // by Greg Ward, p. 88
      SbVec3f diff = *icp.p - ip->p;
      float diff_len = diff.length();

      /*
      // same point?
      if (diff_len == 0.0){ // consider <1e-7
	// check normal, for thin surfaces
	if ( (*icp.n).dot(ip->n) >0.95){
	  *icp.e = ip->irrad;
	  icp.denom = 1.0;
	  icp.num_vals = 1;
	  icp.exactFound = 1;
	  return;
	}
      }
      */

      // wi > 1.0/a
      // 1.0/(e1+e2) > 1.0/a
      // BREAK IF:
      //     e1+e2 < a
      //     e1=|p-pi|/ri   e2=sqrt(1-n.ni)
      //
      //     e1^2 + e2^2 + 2*e1*e2 < a^2
      //     e1^2 < a^2 - e2^2 - 2*e1*e2
      //     e2m=
      
      //     e1* = |p-pi|^2/ri^2 = e1^2
      //     e1^2 > a^2*1.21

      float e1, e2;
      float wi;

//       float wi = 1.0 / (
// 	diff_len / ip->ri +
// 	sqrt( 1.0000004 - (*icp.n).dot(ip->n) ) // sometimes n.n = 1.00000016
// 	);


      // radius test
      e1 = diff.dot(diff)/(ip->ri*ip->ri);
      if (e1 > a*a*1.21){
	if (DEBUG_STATE)
	  cerr << "e1 " << e1 << " too far\n";
	continue;
      }


      // direction test
      float d1 = (*icp.ng).dot(ip->ng);
      float d2 = (*icp.n).dot(ip->ng);   // perturbed normal
      //float d3 = (*icp.n).dot(ip->n);    // perturbed normal
      //float d4 = (*icp.n).dot(ip->ng);   // perturbed normal
      float d = MAX2(d1,d2);
      //e2 = 1.0000004 - (*icp.n).dot(ip->n);
      e2 = 1.00001 - d;

      if (e1+e2 >a*a*1.21){
	if (DEBUG_STATE)
	  cerr << "e2 " << e2 << " too different\n";
	continue;
      }

      // check for stored point in front of current point.
      // we use the actual normal for this calculation (ng)
      // sometimes this is too restrictive, so we also allow
      // points that satisfy the calculation with the interpolated normal.
      SbVec3f diff2 = diff;
      diff.normalize();
      float di = diff.dot( (*icp.n + ip->n) )/2.0;
      float di2 = diff.dot( (*icp.ng + ip->ng) )/2.0;

      // ray behind test
      // do not do in 2d case
      if (gv->useDimensions >= 3){
	
	// di >= 0   cos(t)>cos(pi/2+eps), eps=pi/500
	//if (di<-0.006 || di2 <-0.006){
	float tol = -minRi*a-0.001;
	if ( //di<tol ||
	  di2 < tol){        // di2 < tol
	  if (0 && DEBUG_STATE)
	    cerr << "di " << di << " behind\n";
	  continue;
	}
      }

      e1 = sqrt(e1);
      e2 = sqrt(e2);
      wi = e1 + e2;

      // jitter last test (if not prepass)
      if (wi >a*0.9 + (icp.prepass?0.0:(0.2*icp.ctx->mydrand48())) ){
	if (DEBUG_STATE)
	  cerr << "jitter test failed\n";
	continue;
      }

      // recalculate with perturbed normal
      if (d != d2){
	e2 = sqrt(1.00001 - d2);
	wi = e1 + e2;
      }
	
	if (wi<1e-3)
	  wi=1e3;
	else
	  wi=1.0/wi;
	  
	  icp.num_vals++;

	  // use irradiance gradients to weight
	  SbVec3f nicn = (ip->ng).cross(*icp.n);
	  //SbVec3f nicn = (ip->n).cross(*icp.ng);
	  float delta = 0.0;
	  delta += nicn.dot(ip->Er);  // rotational gradient
	  delta += diff2.dot(ip->Et); // translational gradient

	  // sometimes the gradient is too extreme and the irrad becomes negative!
	  // this is fixed before getIrradiance() returns.
	  
	  *icp.e += wi*( ip->irrad + SbVec3f(delta, delta, delta) );
//	  *icp.e += wi*ip->irrad; // no gradient information
	  icp.denom += wi;
	  
	  if (DEBUG_STATE){
	    cerr << "p:"; printVec(ip->p);
	    cerr << "e:"; printVec(ip->irrad);
	    cerr << "er:"; printVec(ip->Er);
	    cerr << "et:"; printVec(ip->Et);
	    cerr << "r:" << ip->ri;
	    cerr << "\n";
	    showPoint(ip->p, diff_len, ip->irrad);
	  }

    }
      

    // loop over children
    // search child if point is within half of child's-size of child's-boundary
    if (currnode->terminal == 1)
      return;

    // loop over children, stopping if exact point is found
    for (int child = 0; child<8 && !icp.exactFound; child++){
      if (currnode->child[child] == NULL)
	continue;

      ic_node *childnode = currnode->child[child];

      // is p within box of size currsize centered at childcenter?
      if ((*icp.p)[X] < childnode->mid[X]+currnode->half_size &&
	  (*icp.p)[X] > childnode->mid[X]-currnode->half_size &&
	  (*icp.p)[Y] < childnode->mid[Y]+currnode->half_size &&
	  (*icp.p)[Y] > childnode->mid[Y]-currnode->half_size &&
	  (*icp.p)[Z] < childnode->mid[Z]+currnode->half_size &&
	  (*icp.p)[Z] > childnode->mid[Z]-currnode->half_size){
	
	findPoints(icp, childnode);
      }
      
    } // end loop over children
    
} // end of findPoints()


//''''''''''WWW''WW'''WW'WWWW''''''''WW'''''''WW'
//           WW       WW WW  W                WW 
//WW W  WWW  WW  WW  WWW WW  W  WW   WW WWW  WWWW
//WW W W WW  WW  WW W WW WWWW  WW W  WW WW W  WW 
//WW W W WW  WW  WW W WW WW    WW W  WW WW W  WW 
// WW   WWW WWWW WW  WWW WW     WW   WW WW W  WW 
//                                               
//'''''''''''''''''''''''''''''''''''''''''''''''
//
// compares samples position and harmonic mean radius
// with nearby samples to see if they are reasonable.
// Necessary because corner samples sometimes leak.
//
int IrradianceCache::validPoint(SbVec3f &n, SbVec3f &x, float r_, SbColor &e){

  // FIXME: this needs to be adjusted
  // because harm_mean calc changed, radius stuff changed, etc.

  // find closest point
  irpoint *irr = NULL;
  float r = r_*a/4.0; // call r/4 "close"
  closestPoint(root, n, x, r, irr);

  // check if point makes any sense
  // see if (dh/r * de/e1) < 5   <---- made up heuristic
  // delta(harm_mean)/dx * delta(irradiance)/irradiance1 < 5
  if (irr){
    float dh = r_ - irr->ri;
    float dx = (x - irr->p).length();

    // check change in harm_mean
    if (ABSF(dh/dx) > 200.0){
    
      n[0] = dh/dx;
      float de;

      // check each spectra for change in irradiance
      for (int i=0; i<3; i++){
	de = e[i]-irr->irrad[i];

	float test = ABSF(de/irr->irrad[i]);
	x[i] = test;
      
	if ( test > 0.1)
	  return(0);
      }

    }
  }

  return(1);

}


//'''''WWW''''''''''''''''''''''''''WW''WWWW''''''''WW'''''''WW'
//      WW                          WW  WW  W                WW 
// WWW  WW   WW    WWWW  WW   WWWW WWWW WW  W  WW   WW WWW  WWWW
//WW W  WW  WW W  WWW   WWWW WWW    WW  WWWW  WW W  WW WW W  WW 
//WW    WW  WW W    WWW WW     WWW  WW  WW    WW W  WW WW W  WW 
// WWW WWWW  WW   WWWW   WWW WWWW   WW  WW     WW   WW WW W  WW 
//                                                              
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// traverse tree to find the closest sample with similar normal
//
void IrradianceCache::closestPoint(ic_node *currnode, SbVec3f &n, SbVec3f &x,
				   float &r, irpoint *&irr){

    if (currnode == NULL)
      return;

    // search children
    // search child if point is within half of child's-size of child's-boundary
    if (currnode->terminal == 1)
      return;

    // loop over children
    for (int child = 0; child<8; child++){

      ic_node *childnode = currnode->child[child];

      if (childnode == NULL)
	continue;

      // is p within box of size currsize centered at childcenter?
      if (x[X] < childnode->mid[X]+childnode->half_size+r &&
	  x[X] > childnode->mid[X]-childnode->half_size-r &&
	  x[Y] < childnode->mid[Y]+childnode->half_size+r &&
	  x[Y] > childnode->mid[Y]-childnode->half_size-r &&
	  x[Z] < childnode->mid[Z]+childnode->half_size+r &&
	  x[Z] > childnode->mid[Z]-childnode->half_size-r){

	// recurse
	closestPoint(childnode, n, x, r, irr);

      }
      
    } // end loop over children

    
    // loop over points at this node
    list<irpoint *>::iterator I = currnode->pntlist.begin();
    list<irpoint *>::iterator end = currnode->pntlist.end();
    
    for (; I != end; I++){
      irpoint *ip = *I;

      SbVec3f diff = x - ip->p;
      float diff_len2 = diff.dot(diff);

      // check distance
      if (diff_len2<r*r){

	// check normal
	if (n.dot(ip->n)>0.9){

	  /*
	  // debugging
	  static int doneit = 0;
	  if (!doneit++){
	    cerr << "Found close point " << sqrt(diff_len2)
		 << ", " << r << "\n";
	    Ray ray1 = Ray(ip->p, ip->n); ray1.show(45);
	    Ray ray2 = Ray(x, n); ray2.show(25);
	  }
	  */
	  
	  if (DEBUG_STATE){
	    cerr << sqrt(diff_len2) << ", " << ip->ri << ", ";
	    printVec(ip->irrad);
	    cerr << "\n";
	  }

	  // save this point, and distance
	  irr = ip;
	  r = sqrt(diff_len2);
	  
	}
      }

    }

} // end of closestPoint()



//''''''''''''WW''WWWW''WW''''''''''''''
//            WW  WW  W                 
// WWWW  WW  WWWW WW  W WW WWWW W   WWWW
//WWW   WWWW  WW  WW  W WW WW WW W WWW  
//  WWW WW    WW  WW  W WW WW WW W   WWW
//WWWW   WWW  WW  WWWW  WW WW WW W WWWW 
//                                      
//''''''''''''''''''''''''''''''''''''''
//
// sets spatial size of cache domain
//
void IrradianceCache::setDims(SbVec3f newmin, float size, int debug){
  first_cell_size = size;
  min = newmin;
  root->half_size = size/2.0;
  root->mid = newmin + SbVec3f(size/2.0, size/2.0, size/2.0);
  treeDebug = debug;
  minRi = first_cell_size/resolution;
}


//''WW''''''WWW''''''''WW'''''''WW'''W'''''''''WW'''''
//  WW       WW        WW       WWW  W         WW     
// WWW  WW   WW   WW  WWWW  WW  W WW W  WW    WWW  WW 
//W WW WWWW  WW  WWWW  WW  WWWW W  WWW WW W  W WW WWWW
//W WW WW    WW  WW    WW  WW   W   WW WW W  W WW WW  
// WWW  WWW WWWW  WWW  WW   WWW W    W  WW    WWW  WWW
//                                                    
//''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// recursive deletes nodes/children of nodes (but not triangles)
//
void IrradianceCache::deleteNode(ic_node *n){
  if (n == NULL)
    return;

  // remove list
  list<irpoint *>::iterator I = n->pntlist.begin();
  for (; I != n->pntlist.end(); I++){
    delete *I;
  }
  
  // delete children
  if (!n->terminal)
    for (int i=0; i<8; i++)
      deleteNode(n->child[i]);

  // delete self
  delete n;
}


//WWWW''''''''''''''''''''''''''''''WW''''''''''''''''''WW''''''''''''
//WW  W                             WW                  WW            
//WW  W  WW   WWW  WW   WWW   WWWW WWWW W WW W WW  WWW WWWW  WW   W WW
//WW  W WWWW WW W WW W  WW W WWW    WW  WW   W WW WW W  WW  WW W  WW  
//WW  W WW   WW   WW W  WW W   WWW  WW  WW   W WW WW    WW  WW W  WW  
//WWWW   WWW  WWW  WW   WW W WWWW   WW  WW    WWW  WWW  WW   WW   WW  
//                                                                    
//''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Deconstructor
//
IrradianceCache::~IrradianceCache(){
  struct timeval start, end;
  cerr << "Deleting irradiance cache tree..." << flush;
  gettimeofday(&start, NULL); // start delay measure

  cerr << "(octree nodes...)" << flush;
  // delete nodes
  deleteNode(root);

  gettimeofday(&end, NULL);   // end delay measure
  cerr << " done. (" << delay(start, end) << "ms)" << endl;
}


//''''''''''''''''''''''WW'''WWW''''''''''''WW'''''''
//                      WW  WW  W           WW       
//W WW  WW   WWWW  WW  WWWW WW     WWW  WWW WWW   WW 
//WW   WWWW WWW   WWWW  WW  WW    W WW WW W WW W WWWW
//WW   WW     WWW WW    WW  WW  W W WW WW   WW W WW  
//WW    WWW WWWW   WWW  WW   WWW   WWW  WWW WW W  WWW
//                                                   
//'''''''''''''''''''''''''''''''''''''''''''''''''''
//
// Resets tree, deleting all nodes
// This is called by rescanScene() to regenerate an octree
// because of changed geometry.
//
void IrradianceCache::resetCache(){

  // clear tree visualization
  showCube(SbVec3f(0.0, 0.0, 0.0), SbVec3f(0.0, 0.0, 0.0),
	   SbVec3f(0.0, 0.0, 0.0), 1);

  // clear tree
  deleteNode(root);

  // reset variables
  min = SbVec3f(0.0, 0.0, 0.0);
  first_cell_size = 1.0;
  num_pnts = 0;
  root = new ic_node;
  for (int i=0; i<8; i++){
    root->child[i] = NULL;
  }
  root->half_size = first_cell_size/2.0;
  root->mid = min + SbVec3f(first_cell_size/2.0,
			    first_cell_size/2.0,
			    first_cell_size/2.0);
  root->terminal = 1;
}
  


