
#ifndef KDTREECLASS
#define KDTREECLASS

#include "Main.h"
#include "Triangle.h"
#include "TriangleList.h"
#include "RayAccel.h"
class Ray; // forward declaration of Ray class

#define EPSILON_OS  0.03    // (KdTree Size)
#define EPSILON_OI  0.00005 // (KdTree Include)
//#define SHOWTRAV 1

class KdTree : public RayAccel {
private:

  enum event_t { END, PLANAR, BEGIN };

  typedef struct split_t {
    float pos;
    int axis;
    split_t(float pos_, int axis_) : pos(pos_), axis(axis_) { }
  } split_t;
  
  // event item
  // this could be smaller
  typedef struct fullEvent_t {
    union {
      float pos;
      unsigned int posui;
    };
    event_t eventType; // last two bits append to above when sorting
    int axis;
    int triId;
    fullEvent_t(){}
    fullEvent_t(float pos_, int axis_, event_t eventType_, int triId_) :
      pos(pos_), axis(axis_), eventType(eventType_), triId(triId_) { }
  } fullEvent_t;
  
  enum listNames { LIST_NULL, LIST_EMPTY, LIST_WHOLE };
  enum nodeNames { NODE_NULL, NODE_ROOT };
  enum sideNames { LEFT, RIGHT };

  // parameters struct
  typedef struct treeParam {
    Ray *r;                 // ray
    SbVec3f invDir;
    const Triangle *ignoreTri;
    unsigned int front[3], back[3];
  } treeParam;

  // cell
  // data structure from wald's thesis
  typedef struct nodeLeaf {
    unsigned int flagIndex;
    // bits 0..30 : list index (listi)
    //         31 : inner=0 or leaf=1
    float pad;
  } nodeLeaf;
  
  typedef struct nodeInner {
    unsigned int flagAxisOffset;
    // bits 0...1 : axis
    //      2..30 : child offset
    //         31 : inner=0 or leaf=1
    float mid;
  } nodeInner;

  typedef union {
    nodeLeaf leaf;
    nodeInner inner;
    
    unsigned int isLeaf()   const { return inner.flagAxisOffset & (unsigned int)(1<<31); }
    unsigned int axis()     const { return inner.flagAxisOffset & (0x3); }
    unsigned int childOfs() const { return inner.flagAxisOffset & 0x7FFFFFFC; }      // absolute (not relative)
    unsigned int childIdx() const { return (inner.flagAxisOffset & 0x7FFFFFFC)>>3    //   memory offset (8*) from nodes[0]
				      ; } // array offset
    unsigned int listIdx()  const { return leaf.flagIndex      & 0x7FFFFFFF; }
    float mid()             const { return inner.mid; }
    void setLeaf() { inner.flagAxisOffset |= (unsigned int)(1<<31); }
    void setInner(){ inner.flagAxisOffset &= 0x7FFFFFFF; }
    void setAxis(unsigned int x)    { inner.flagAxisOffset = (inner.flagAxisOffset & 0xFFFFFFFC) | x; }
    void setChildOfs(unsigned int x){ inner.flagAxisOffset = (inner.flagAxisOffset & 0x80000003) | x; }
    void setChildIdx(unsigned int x){ inner.flagAxisOffset = (inner.flagAxisOffset & 0x80000003) | (x<<3); }
    void setListIdx(unsigned int x) { leaf.flagIndex      = (leaf.flagIndex      & 0x80000000) | x; }
    float setMid(float x)           { inner.mid = x; }
//     int li;                  // 4 bytes, list index
//     int axis;                // 4 bytes
//     int childi;              // 4 bytes, child index
//     float mid;               // 4 bytes
  } node;

#define NODE_ISLEAF(n)      ((n)->inner.flagAxisOffset & (unsigned int)(1<<31))
#define NODE_DIMENSION(n)   ((n)->inner.flagAxisOffset & (0x3))
#define NODE_CHILDOFFSET(n) ((n)->inner.flagAxisOffset & (0x7FFFFFFC))
#define NODE_LISTINDEX(n)   ((n)->leaf.flagIndex       & (0x7FFFFFFF))

  
//   typedef struct node_st {
//     int li;                  // 4 bytes, list index
//     int axis;                // 4 bytes
//     int childi;              // 4 bytes, child index
//     float mid;               // 4 bytes

// #ifdef SHOWTRAV  
//     // debugging
//     SbVec3f min, max;
//     int depth;
// #endif
//  } node;
  
  //static vector<stackElem_t> kdstack;
  typedef struct stackElem2 {
    node *n;  // node index
    int pad;
    double t0, t1;
  } stackElem2_t;
  static stackElem2_t stk[64];
 
  int kdtreeDebug;
  int maxItems;
  int maxDepth;
  node *root;
  int rooti;
  node *compactNodes;
  vector<node> nodes;
  vector<TriangleList> lists;
  //TriangleList tri_list;
  unsigned int rayId;
  pthread_mutex_t lock;
  int reordered;

  int findSplit(int ni, float vmin[3], float vmax[3],  // ni == node index
		int &a, float &p,
		TriangleList &leftList, TriangleList &rightList);

  int findSplitFast(int ni, // ni == node index
		    int niNumTris,
		    const float vmin[3], const float vmax[3],  
		    vector <fullEvent_t> &eventList,
		    int &min_a, float &min_p, int &min_leftOrRight,
		    int &min_left, int &min_right, int &min_planar);
  
  void refineNode(int ni, int depth, int simpleDiv,
		  float vmin[3], float vmax[3]);
  void sortAndRefine(int ni, int depth, const float vmin[3],
		     const float vmax[3]);
  void fastRefineNode(vector<fullEvent_t> *eventList, int ni,
		      int niNumTris, int depth, 
		      const float vmin[3], const float vmax[3]);
  void radixSort(vector<fullEvent_t> &orig, int numVals);
  
  int proc_subtree(const double &t0, const double &t1, const int ni,
		   const treeParam &tp) const;
  
  float SAH(split_t p, const float vmin[3], const float vmax[3],
	    int nl, int nr, int np, int &leftOrRight);

  void printEventList(vector <fullEvent_t> *eventList);
  void printBitVector(vector<bool> *vec);


  // temporary!
  int rays;

//   int num_tris;
//   float min[3], max[3];
//   double maxSideLength;

//   // keep bounding sphere of triangles
//   SbVec3f center;
//   float radius;

//   // also keep bounding sphere of specular triangles
//   SbVec3fshiny_center;
//   float shiny_radius;

  SbVec3f shiny_min, shiny_max;
  
  void deleteNode(int ni);
  void printStats(int ni, int depth=0);
  int ss_intersect_ray(Ray &r);

  friend ostream& operator<< (ostream& os, const fullEvent_t &e);
  
public:


  KdTree();
  ~KdTree();
  void reset();

  int add(Triangle *t){

    // add triangle
    num_tris++;
    return lists[LIST_WHOLE].add(t);    
  }

  void printStats(){ printStats(NODE_NULL, 0); }
  void setParams(int maxD, int maxI, int debug);
  void refine(int simpleDiv = 0);
  int intersect_ray(Ray &r);
  void reorderNodes();

//   int size(){ return(num_tris); }
//   double& getMaxSideLength(){ return maxSideLength; } 
//   float* getMin(){ return min; }
//   float* getMax(){ return max; }
//   SbVec3f& getCenter(){ return center; }
//   float& getRadius(){ return radius; }
//   float& getShinyRadius(){ return shiny_radius; }
//   SbVec3f& getShinyCenter(){ return shiny_center; }

};

#endif /* KDTREECLASS */
