////////////////////////////////////////////////////////////////////////////
//
//WWWW''WW''''''''''WW''''''''''''''''''''''''''''''''''''''''''''WWW'''''''
//WW  W WW          WW                                             WW       
//WW  W WWW   WW   WWWW  WW   WWW       WWWW W   WWW WWW      WWW  WW   WWW 
//WWWW  WW W WW W   WW  WW W  WW W      WW WW W W WW WW W    WW W  WW  W WW 
//WW    WW W WW W   WW  WW W  WW W      WW WW W W WW WW W    WW    WW  W WW 
//WW    WW W  WW    WW   WW   WW W WWWW WW WW W  WWW WWW      WWW WWWW  WWW 
//                                                   WW                     
//'''''''''''''''''''''''''''''''''''''''''''''''''''WW'''''''''''''''''''''
//
// File: Photon_map.cxx
// Description: Photon_map class implementation
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// Photon map class for the Pane ray tracer
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// class Photon_map implements the photon map
// data structure as described by Henrik Jensen
// in the Photon Mapping algorithm described
// in his book:
//
// "Realistic Image Synthesis
// Using Photon Mapping" by Henrik Wann Jensen 2001
// http://graphics.ucsd.edu/~henrik/papers/book/
//
// Briefly, "photons" (particles representing radiant
// flux) are stored along with their position and
// incoming direction in a kd-tree data structure,
// which, when balanced, provides fast nearest-neighbor
// searches.
//
// Using this nearest-neighbor search, a small
// neighborhood of photons can be found, and using
// the area they subtend an approximate estimate
// of the irradiance can be calculated.
//
// This source code is based on an implementation
// from that book. However, it has been extended a
// fair amount.
// 
// Differences:
//   * irradiance can be precomputed
//   * automatic maximum search radius 
//   * irradiance estimate uses cone filter
//   * photons store normal and precomputed irradiance
//   * photon power and irradiance are compressed, so
//     photon size is actually 2 bytes smaller (26 vs. 28).
//   * "bug" fixed in locate_photons to find exact nearest
// 
// $Id: Photon_map.cxx 1186 2007-11-12 09:02:06Z beason $
//
////////////////////////////////////////////////////////////////////////////
 
#include "Photon_map.h"

// This is the constructor for the photon map.
// To create the photon map it is necessary to specify the
// maximum number of photons that will be stored
//
//=================================================================
Photon_map::Photon_map( const int max_phot, GlobalVars *gv_ptr )
//=================================================================
{
  stored_photons = 0;
  // prev_scale = 1;
  max_photons = max_phot;
  gv = gv_ptr;

  const unsigned long size = sizeof( Photon ) * ( max_photons+1 );
  cerr << "Allocating " << size << " bytes for photon map..." << endl;
  photons = (Photon*)malloc( size );
  //photons = new Photon[max_photons+1];

  if (photons == NULL) {
    fprintf(stderr,"Out of memory initializing photon map\n");
    exit(-1);
  }

  bbox_min[0] = bbox_min[1] = bbox_min[2] = 1e99f;
  bbox_max[0] = bbox_max[1] = bbox_max[2] = -1e99f;

  //----------------------------------------
  // initialize direction conversion tables
  //----------------------------------------

  for (int i=0; i<256; i++) {
    double angle = double(i)*(1.0/256.0)*M_PI;
    costheta[i] = cos( angle );
    sintheta[i] = sin( angle );
    cosphi[i]   = cos( 2.0*angle );
    sinphi[i]   = sin( 2.0*angle );
  }

  // initialize buffers
  for (int i=0; i<maxThreads; i++){
    buffers[i] = NULL;
    buffPos[i] = 0;
  }

  half_stored_photons = 0;

  bound_rad = bound_rad2 = 0.0;
  bbox_cen = SbVec3f(0,0,0);
  useCompression = false;
  sphereCompression = 1.0;
}

// Deconstructor
//
//==========================
Photon_map::~Photon_map()
//==========================
{
  free( photons );
  //delete [] photons;

  for (int i=0; i<maxThreads; i++)
    free( buffers[i] );
}

// photon_dir returns the direction of a photon
//
//===================================================================
void Photon_map::photon_dir( float *dir, const Photon *p ) const
//===================================================================
{
  dir[0] = sintheta[p->theta]*cosphi[p->phi];
  dir[1] = sintheta[p->theta]*sinphi[p->phi];
  dir[2] = costheta[p->theta];
}

// photon_norm returns the direction of the surface normal at the photon
//
//===================================================================
void Photon_map::photon_norm( float *norm, const Photon *p ) const
//===================================================================
{
  norm[0] = sintheta[p->normt]*cosphi[p->normp];
  norm[1] = sintheta[p->normt]*sinphi[p->normp];
  norm[2] = costheta[p->normt];
}

float Photon_map::max_radius(const int nphotons) const {
  const float r_alpha = 0.1;
  float max = sqrt(nphotons*bound_rad2/
                   (r_alpha * (stored_photons ? stored_photons:1)));
  const static bool largeRad = (bool)getenv("PANE_LARGE_RADIUS");
  if (largeRad){
    max = MAX2(max, bound_rad);
    cerr << "Using LARGE RADIUS of " << max << endl;
  }
  return max;
}


// store_photon_floats compresses a float array into photon as four bytes
//
// Power compression is from
// Real Pixels, Graphics Gems II, p.80
// by Greg Ward
//
//==================================================================
void Photon_map::store_photon_floats( const float *pow,
					unsigned char *p ) const
//==================================================================
{
  float v = MAX3(pow[R], pow[G], pow[B]);

  if (v<1e-32f)
    p[R] = p[G] = p[B] = p[E] = 0;
  else {
    int e;
    v = frexp(v, &e)*256./v;
    p[R] = (unsigned char)(pow[R]*v);
    p[G] = (unsigned char)(pow[G]*v);
    p[B] = (unsigned char)(pow[B]*v);
    p[E] = e+128;
  }
}

// get_photon_floats uncompresses photon irrad or power into float array
//
//===================================================================
void Photon_map::get_photon_floats( float *pow,
				      const unsigned char *p ) const
//===================================================================
{
  if (p[E] == 0)
    pow[R] = pow[G] = pow[B] = 0.0;
  else {
    float v = ldexp(1./256., p[E]-128);
    pow[R] = (p[R]+0.5)*v;
    pow[G] = (p[G]+0.5)*v;
    pow[B] = (p[B]+0.5)*v;
  }
}

// irradiance_estimate computes an irradiance estimate
// at a given surface position
//
//======================================
void Photon_map::irradiance_estimate(
  float irrad[3],                // returned irradiance
  const float pos[3],            // surface position
  const float normal[3],         // surface normal at pos
  const int nphotons ) const     // number of photons to use
//======================================
{
  irrad[0] = irrad[1] = irrad[2] = 0.0;
  if (stored_photons==0) return;
  
  NearestPhotons np;
  np.dist2 = (float*)alloca( sizeof(float)*(nphotons+1) );
  np.index = (const Photon**)alloca( sizeof(Photon*)*(nphotons+1) );

  np.pos[0] = pos[0];
  np.pos[1] = pos[1];
  np.pos[2] = pos[2];
  np.check_norm = 0;
  np.check_from_diffuse = 0;
  np.max = nphotons;
  np.found = 0;
  np.got_heap = 0;
  np.dist2[0] = max_dist2_pre;
  np.norm[0] = normal[0];
  np.norm[1] = normal[1];
  np.norm[2] = normal[2];
//  np.dist2[0] = 1.0e99;

  // for compressing sphere into disc
  //np.sbr.setValue(SbVec3f(0.0, 1.0, 0.0), SbVec3f(normal));
  //np.sbr.setValue(SbVec3f(normal), SbVec3f(0.0, 1.0, 0.0));

  // locate the nearest photons
  locate_photons( &np, 1 );

  // debugging
//   cerr << "Found " << np.found << " photons."
//        << " stored_photons=" << stored_photons
//        << " half_stored_photons=" << half_stored_photons
//        << "\n";

  // if less than 8 photons return
  if (np.found<8)
    return;

  // photon filtering from jensen p.82
  const float gauss_norm = 1.8790;            // normalization of gaussian
  const float alpha = 0.918 * gauss_norm;
  const float beta = 1.953;
  float wpg;
  
//   const float k = 1.1;
//   const float kdiv = (1.0-2.0/(3.0*k));
//   float wpc;
//   float r = sqrt(np.dist2[0]);
  
  float pdir[3];
  float pow[3];

  // temp
  float ndir[3];
  // float ymin=1.0e99, ymax=-1.0e99;
  
  // sum irradiance from all photons
  for (int i=1; i<=np.found; i++) {
    const Photon *p = np.index[i];
    // the photon_dir call and following if can be omitted (for speed)
    // if the scene does not have any thin surfaces
    photon_dir( pdir, p );

    if ( (pdir[0]*normal[0]+pdir[1]*normal[1]+pdir[2]*normal[2]) < 0.0f ) {

    // debugging
    //Ray photondir(SbVec3f(p->pos),SbVec3f(pdir));
    //photondir.show(2.0);
    
      // photon filtering

      // tent filter
//       float dp = sqrt(np.dist2[i]);
//       wpc = 1.0-dp/(k*r);

      // gaussian filter
      wpg =

	alpha * (
	  
	1.0 -

	( 1.0 -
	  exp( -beta * np.dist2[i] / (2.0*np.dist2[0]) )
	  )
	
	/
	
	( 1.0 - exp(-beta) )
	
	);

      get_photon_floats(pow, p->power);

      //debugging
      //cerr << "photon: "; printVec(SbVec3f(pow)); cerr << "\n";

      // 2d adjustment
      if (gv->useDimensions<3){

// 	// record f min and max for sphere clipping
// 	if (p->pos[1]>ymax)
// 	  ymax = p->pos[1];
// 	if (p->pos[1]<ymin)
// 	  ymin = p->pos[1];
	
	photon_norm( ndir, p);
	if (ndir[1]<1.0)
	  // wpg = wpg * 1/cos(angle with plane y=0)
	  wpg /= sqrt(1.0-ndir[1]*ndir[1]);
      }

//      wpg = wpc;
//      wpg = 1.0;
      irrad[0] += pow[0]*wpg;
      irrad[1] += pow[1]*wpg;
      irrad[2] += pow[2]*wpg;
    }
  }
  

  
  //const float wpc_norm = (1.-2./(3.*k));
  const float wpc_norm = 1.0;
  float tmp;

//  if (gv->useDimensions==3)
    tmp = (1.0f/M_PI)/( wpc_norm * np.dist2[0]);  // estimate of density
//   else {
//     float ydiff = ymax-ymin;
//     tmp = 1.0f/(sqrt(np.dist2[0])*2.0*ydiff);;
//   }


  irrad[0] *= tmp;
  irrad[1] *= tmp;
  irrad[2] *= tmp;
}


// locate_photons find the nearest photons in the
// photon map given the parameters in np
//
//===========================================
void Photon_map::locate_photons(
  NearestPhotons *const np,
  const int index ) const
//===========================================
{
  const Photon *p = &photons[index];
  float dist1;

  //if (index<half_stored_photons) {
  // experimenting 2/4/05
  if (index<=half_stored_photons) {
    dist1 = np->pos[ p->plane ] - p->pos[ p->plane ];

    if (dist1>0.0) { // if dist1 is positive search right plane
      locate_photons( np, 2*index+1 );
      if ( dist1*dist1 < np->dist2[0] )
	locate_photons( np, 2*index );
    } else {         // dist1 is negative search left first
      locate_photons( np, 2*index );
      if ( dist1*dist1 < np->dist2[0] )
	locate_photons( np, 2*index+1 );
    }
  }


//   // compress sphere into disc
//   SbVec3f dir(p->pos[0]-np->pos[0],
// 	      p->pos[1]-np->pos[1],
// 	      p->pos[2]-np->pos[2]);
//   SbVec3f newdir;
//   np->sbr.multVec(dir, newdir);
//   newdir[1] *= 3.0;          // compress Y axis
//   float dist2 = newdir.dot(newdir);


  // compute squared distance between current photon and np->pos
  dist1 = p->pos[0] - np->pos[0];
  float dist2 = dist1*dist1;
  dist1 = p->pos[1] - np->pos[1];
  dist2 += dist1*dist1;
  dist1 = p->pos[2] - np->pos[2];
  dist2 += dist1*dist1;

  // compress sphere into disc
  if (useCompression){

    //   y| .   r^2 = x^2 + y^2 ==> r^2 = x^2 + (y/c)^2
    //____|____                     where x^2 = p^2 - y^2
    //    | x                       and   p^2 = dist2
    //    |                         and   y^2 = n.(p-p0)^2
    //          r^2 = dist2 + (1/c^2-1)*n.(p-p0)^2
    
    float ndotp =      // n.(p-p0)
      np->norm[0]*(p->pos[0]-np->pos[0]) +
      np->norm[1]*(p->pos[1]-np->pos[1]) +
      np->norm[2]*(p->pos[2]-np->pos[2]);

    dist2 += sphereCompression*ndotp*ndotp;
  }

  if ( dist2 < np->dist2[0] ) {


//     // debugging
//     static int num = 0;
//     if (DEBUG_STATE && num++<100000){
//       float norm[3];
//       photon_norm(norm, p);
//       Ray r(SbVec3f(p->pos), SbVec3f(norm));    // photon normal
//       r.show2(20.0);
//       Ray r2(SbVec3f(p->pos), SbVec3f(np->norm));  // surface normal
//       r2.show2(10.0);
//       //showPoint(SbVec3f(np->pos), 10.0);
//       showPhoton(p->pos);
//
//       // print photon power
//       float power[3];
//       get_photon_floats( power, p->power );
//       cerr << "photon position: "; printVec(SbVec3f(p->pos)); cerr << " ";
//       cerr << "photon power: "; printVec(SbVec3f(power)); cerr << endl;
//      
//     }


    // check that photon is indirect
    if (np->check_from_diffuse && !p->from_diffuse)
      return;
    
    // check for "similar normal"
    if (np->check_norm){
      float norm[3];
      photon_norm(norm, p);

      if ((norm[0]*np->norm[0]+norm[1]*np->norm[1]+norm[2]*np->norm[2]) < 0.9f)
	return;

    }

    // we found a photon :) Insert it in the candidate list
    if ( np->found < np->max ) {
      // heap is not full; use array
      np->found++;
      np->dist2[np->found] = dist2;
      np->index[np->found] = p;
    } else {
      int j,parent;

      if (np->got_heap==0) { // Do we need to build the heap?
	
	// Build heap
	float dst2;
	const Photon *phot;

	// builds heap by looping over parent nodes and enforcing
	// heap condition downwards
	int half_found = np->found>>1;
	for ( int k=half_found; k>=1; k--) { 
	  parent=k;
	  phot = np->index[k];
	  dst2 = np->dist2[k];

	  // fixDown() from Sedgewick
	  // enforces heap condition down the tree
	  while ( parent <= half_found ) {
	    j = parent+parent;      // j = child
	    if (j<np->found && np->dist2[j]<np->dist2[j+1]) // larger child
	      j++;
	    if(dst2>=np->dist2[j])
	      break;
	    np->dist2[parent] = np->dist2[j];   // shifts child up tree
	    np->index[parent] = np->index[j];
	    parent=j;
	  }
	  np->dist2[parent] = dst2;   // moves parent to bottom of shifted path
	  np->index[parent] = phot;
	}
	np->got_heap = 1;

	// special case where new photon is too far away
	// added by kbeason 2/4/05
	if (dist2 > np->dist2[1]){
	  np->dist2[0] = np->dist2[1];
	  return;
	}
      }

      // insert a new photon into max heap
      // delete largest element, insert new, and reorder the heap
      parent=1;
      j = 2;
      while ( j <= np->found ) {
	if ( j < np->found && np->dist2[j] < np->dist2[j+1] )
	  j++;
	if ( dist2 > np->dist2[j] )
	  break;
	np->dist2[parent] = np->dist2[j];
	np->index[parent] = np->index[j];
	parent = j;
	j += j;
      }

      // check moved up by kbeason 2/4/05
      // check added by kbeason 8/5/03
      // important for finding nearEST photons
      // otherwise one of the photons may get overwritten
      // by a photon that is closer than dist2[0]
      // but not closer than max(dist2[i>0])
      //if (dist2 < np->dist2[parent]){
      np->index[parent] = p;
      np->dist2[parent] = dist2;
      //}
      
      np->dist2[0] = np->dist2[1];
    } // end if (np->found < np->max )
  } // end  if ( dist2 < np->dist2[0] )
}


namespace {
  pthread_mutex_t lock_store = PTHREAD_MUTEX_INITIALIZER;
}

// buffered_store puts a photon into
// a buffer, periodically placing
// the entire buffer into the flat array that will form
// the final kd-tree. 
// Faster than store() when multi-threading.
//
// Call this function to store a photon.
//
//=================================
void Photon_map::buffered_store(
//=================================
  const float power[3],
  const float pos[3],
  const float dir[3],
  const float normal[3],
  const int light_id,
  const int from_diffuse,
  const int thread,
  const int flush)
{

  const int bufferSize = 10000;
  if (thread>maxThreads){
    cerr << "Too many threads in Photon_map!\n";
    return;
  }

  // make buffer
  if (!buffers[thread])
    buffers[thread] = (Photon *)malloc( sizeof( Photon ) * bufferSize );

  // flush buffer?
  if (flush || buffPos[thread] == bufferSize){

    // copy buffer into map, photon by photon
    pthread_mutex_lock(&lock_store);
    for (int i=0; i<buffPos[thread]; i++){

      // check to see if map is full
      if (stored_photons>=max_photons){
	pthread_mutex_unlock(&lock_store);
	return;
      }

      photons[++stored_photons] = buffers[thread][i];

      // debugging
      //cerr << "INSERT: stored_photons=" << stored_photons
      //   << " max_photons=" << max_photons << "\n";
    }
    pthread_mutex_unlock(&lock_store);

    buffPos[thread] = 0;
    
    // quit if we are just flushing
    if (flush)
      return;
  }

  // store photon into buffer
  store(power, pos, dir, normal, light_id, from_diffuse,
	&buffers[thread][buffPos[thread]++]);
}

// store puts a photon into the flat array that will form
// the final kd-tree.
//
// Call this function to store a photon.
//
//===============================
void Photon_map::store(
  const float power[3],
  const float pos[3],
  const float dir[3],
  const float normal[3],
  const int light_id,
  const int from_diffuse,
  Photon *buffer)        // if buffer is supplied, store photon in there
//===============================
{

  int buffer_idx = 0;

  // if no buffer is given, store in map.
  // buffer is optional argument for multi-threading:
  // we may store in a buffer to prevent
  // the blocking that comes from storing in the map.
  if (buffer == NULL){
    buffer = photons;

    pthread_mutex_lock(&lock_store);

    // check to see if map is full
    if (stored_photons>=max_photons){
      pthread_mutex_unlock(&lock_store);
      return;
    }

    // create new photon
    buffer_idx = ++stored_photons;
    pthread_mutex_unlock(&lock_store);
  }

  Photon *const node = &buffer[buffer_idx];

  for (int i=0; i<3; i++) {
    node->pos[i] = pos[i];

    if (node->pos[i] < bbox_min[i])
      bbox_min[i] = node->pos[i];
    if (node->pos[i] > bbox_max[i])
      bbox_max[i] = node->pos[i];
  }
  store_photon_floats(power, node->power);

  // store photon direction
  int theta = int( acos(dir[2])*(256.0/M_PI) );
  if (theta>255)
    node->theta = 255;
  else
    node->theta = (unsigned char)theta;

  int phi = int( atan2(dir[1],dir[0])*(256.0/(2.0*M_PI)) );
  if (phi>255)
    node->phi = 255;
  else if (phi<0)
    node->phi = (unsigned char)(phi+256);
  else
    node->phi = (unsigned char)phi;

  // store surface normal
  int normt = int( acos(normal[2])*(256.0/M_PI) );
  if (normt>255)
    node->normt = 255;
  else
    node->normt = (unsigned char)normt;

  int normp = int( atan2(normal[1],normal[0])*(256.0/(2.0*M_PI)) );
  if (normp>255)
    node->normp = 255;
  else if (normp<0)
    node->normp = (unsigned char)(normp+256);
  else
    node->normp = (unsigned char)normp;

  // store light id and from_diffuse
  node->light_id = light_id;
  node->from_diffuse = from_diffuse;
}


//  scale_photon_power is used to scale the power of all
//  photons once they have been emitted from the ligh
//  source. scale = 1/(#emitted photons).
//  Call this function after each light source is processed.
//
//=========================================================
void Photon_map::scale_photon_power( const float scale )
//=========================================================
{
  float power[3];
  for (int i=1; i<=stored_photons; i++) {
    get_photon_floats(power, photons[i].power);
    power[0] *= scale;
    power[1] *= scale;
    power[2] *= scale;
    store_photon_floats(power, photons[i].power);
  }
}


// balance creates a left-balanced kd-tree from the flat photon array.
// This function should be called before the photon map
// is used for rendering.
//
//===============================
void Photon_map::balance(const int nphotons)
//===============================
{
  if (stored_photons>1) {
    // allocate two temporary arrays for the balancing procedure
    Photon **pa1 = (Photon**)malloc(sizeof(Photon*)*(stored_photons+1));
    Photon **pa2 = (Photon**)malloc(sizeof(Photon*)*(stored_photons+1));

    for (int i=0; i<=stored_photons; i++)
      pa2[i] = &photons[i];

    balance_segment( pa1, pa2, 1, 1, stored_photons );
    free(pa2);

    // reorganize balanced kd-tree (make a heap)
    int d, j=1, foo=1;
    Photon foo_photon = photons[j];

    for (int i=1; i<=stored_photons; i++) {
      d=pa1[j]-photons;
      pa1[j] = NULL;
      if (d != foo)
	photons[j] = photons[d];
      else {
	photons[j] = foo_photon;

	if (i<stored_photons) {
	  for (;foo<=stored_photons; foo++)
	    if (pa1[foo] != NULL)
	      break;
	  foo_photon = photons[foo];
	  j = foo;
	}
	continue;
      }
      j = d;
    }
    free(pa1);
  }

  // debugging
//   int c[2] = {0};
//   float power[3];
//   for (int i=0; i<=stored_photons; i++){
//     get_photon_floats( power, photons[i].power );
//     if (power[0]>power[2])
//       c[0]++;
//     else
//       c[1]++;
//   }
//   cerr << "\nPhoton storage results:\n"
//        << c[0] << " red\n"
//        << c[1] << " blue\n";

  
  //half_stored_photons = stored_photons/2-1;
  // experimenting 2/4/05
  half_stored_photons = (stored_photons-1)/2;

  // reasonable defaults for empty map
  if (!stored_photons){
    bbox_max[0] = bbox_min[0] = 0.0;
    bbox_max[1] = bbox_min[1] = 0.0;
    bbox_max[2] = bbox_min[2] = 0.0;
  }
  
  // find radius of bounding sphere
  bbox_cen[0] = (bbox_max[0]+bbox_min[0])*0.5;
  bbox_cen[1] = (bbox_max[1]+bbox_min[1])*0.5;
  bbox_cen[2] = (bbox_max[2]+bbox_min[2])*0.5;

//   cerr << "Bounding box min:"
//        << bbox_min[0] << "," << bbox_min[1] << "," << bbox_min[2] << endl;
//   cerr << "Bounding box max:"
//        << bbox_max[0] << "," << bbox_max[1] << "," << bbox_max[2] << endl;

  // compute max-cen length
  float dist;
  dist = bbox_max[0]-bbox_cen[0];
  bound_rad2 = dist*dist;
  dist = bbox_max[1]-bbox_cen[1];
  bound_rad2 += dist*dist;
  dist = bbox_max[2]-bbox_cen[2];
  bound_rad2 += dist*dist;

  bound_rad = sqrt(bound_rad2);


  // precompue maximum search distance
  max_dist2_pre = max_radius(nphotons);
  max_dist2_pre *= max_dist2_pre;
  
  // debugging
  cerr << "Bounding radius (squared) is " << bound_rad2 << "\n";
  
}

// swaps two photons in array
#define swap(ph,a,b) { Photon *ph2=ph[a]; ph[a]=ph[b]; ph[b]=ph2; }

// median_split splits the photon array into two separate
// pieces around the median, with all photons below
// the median in the lower half and all photons above
// the median in the upper half. The comparison
// criteria is the axis (indicated by the axis parameter)
// (inspired by routine in "Algorithms in C++" by Sedgewick
//================================================================
void Photon_map::median_split(
  Photon **p,
  const int start,              // start of photon block in array
  const int end,                // end of photon block in array
  const int median,             // desired median number
  const int axis )              // axis to split along
//================================================================
{
  int left = start;
  int right = end;

  while ( right > left ) {
    const float v = p[right]->pos[axis];
    int i=left-1;
    int j=right;
    for (;;) {
      while ( p[++i]->pos[axis] < v )
	;
      while ( p[--j]->pos[axis] > v && j>left )
	;
      if ( i >= j )
	break;
      swap(p,i,j);
    }

    swap(p,i,right);
    if ( i >= median )
      right=i-1;
    if ( i <= median )
      left=i+1;
  }
}


// See "Realistic Image Synthesis using Photon Mapping" Chapter 6
// for an explanation of this function
//
//==================================
void Photon_map::balance_segment(
  Photon **pbal,
  Photon **porg,
  const int index,
  const int start,
  const int end )
//==================================
{
  //--------------------
  // compute new median (such that tree about it will be complete)
  //--------------------

  int median=1;
  while ((4*median) <= (end-start+1))
    median += median;

  if ((3*median) <= (end-start+1)) {
    median += median;
    median += start-1;
  } else
    median = end-median+1;

  //--------------------------
  // find axis to split along
  //--------------------------

  int axis=2;
  if ((bbox_max[0]-bbox_min[0])>(bbox_max[1]-bbox_min[1]) &&
      (bbox_max[0]-bbox_min[0])>(bbox_max[2]-bbox_min[2]))
    axis=0;
  else if ((bbox_max[1]-bbox_min[1])>(bbox_max[2]-bbox_min[2]))
    axis=1;

  //------------------------------------------
  // partition photon block around the median
  //------------------------------------------

  median_split( porg, start, end, median, axis );

  pbal[ index ] = porg[ median ];
  pbal[ index ]->plane = axis;

  //----------------------------------------------
  // recursively balance the left and right block
  //----------------------------------------------

  if ( median > start ) {
    // balance left segment
    if ( start < median-1 ) {
      const float tmp=bbox_max[axis];
      bbox_max[axis] = pbal[index]->pos[axis];
      balance_segment( pbal, porg, 2*index, start, median-1 );
      bbox_max[axis] = tmp;
    } else {
      pbal[ 2*index ] = porg[start];
    }
  }

  if ( median < end ) {
    // balance right segment
    if ( median+1 < end ) {
      const float tmp = bbox_min[axis];
      bbox_min[axis] = pbal[index]->pos[axis];
      balance_segment( pbal, porg, 2*index+1, median+1, end );
      bbox_min[axis] = tmp;
    } else {
      pbal[ 2*index+1 ] = porg[end];
    }
  }
}


// precompute the irradiance and store at each photon.
// spawns threads calling precompute_irradiance_sec()
// returns average max distance of photons in estimate
//
//==============================================================
float Photon_map::precompute_irradiance(const int nphotons,
					  const int num_threads,
					  const int show)
//==============================================================
{

  // clear photon display
  if (show)
    showPhoton(SbVec3f(0.0, 0.0, 0.0), SbColor(0.0, 0.0, 0.0), 1);
    
  pthread_t threads[maxThreads] = {0};
  pmMemberFunction memberFunc[maxThreads]; // used to pass object pointer

  // create threads
  for (int i=0; i<num_threads; i++){
    
      // pack object pointer, function pointer, and argument into struct
      memberFunc[i].object = this;
      memberFunc[i].pmf = &Photon_map::precompute_irradiance_section;
      memberFunc[i].nphotons = nphotons;
      memberFunc[i].thread = i;
      memberFunc[i].num_threads = num_threads;
      memberFunc[i].show = show;
      memberFunc[i].avgdist2 = 0.0;

      // call wrapper function with struct
      pthread_create(&threads[i], NULL,
		     Photon_map::precompute_wrapper,
		     (void*)&memberFunc[i]);
  }

  // wait for threads to finish
  float avgdist2 = 0.0;
  for (int i=0; i<num_threads; i++){
    pthread_join(threads[i], NULL);
    avgdist2 += memberFunc[i].avgdist2;
  }

  // compute average distance of estimate
  avgdist2 /= num_threads;
  return(sqrt(avgdist2));
}

// precompute irradiance for a subset of photons
// See http://www.acm.org/jgt/papers/Christensen99/
//
// For each photon, computes radiance estimate
// and stores estimate inside photon.
//
//======================================================================
void Photon_map::precompute_irradiance_section(float &avgdist2,
						 const int nphotons,
						 const int thread,
						 const int num_threads,
						 const int show
						 )
//======================================================================
{
  // check to see if any photons
  if (stored_photons == 0)
    return;

  // photon filtering from jensen p.82
  const float k = 1.1;
  const float kdiv = (1.0-2.0/(3.0*k));
  float wpc;
  avgdist2 = 0.0;
  int avgdisti = 0;
  
  // precompute radiance
  Photon *node;
  NearestPhotons np;
  np.dist2 = (float*)alloca( sizeof(float)*(nphotons+1) );
  np.index = (const Photon**)alloca( sizeof(Photon*)*(nphotons+1) );

  for (int i=thread+1; i<=stored_photons; i+=num_threads){

    // status message
    if (i%10000==0){
      fprintf(stderr, "\r%6.2f%%", (float)i*100.0/stored_photons);
      fflush(stderr);
    }
    
    node = &photons[i];

    np.pos[0] = node->pos[0];
    np.pos[1] = node->pos[1];
    np.pos[2] = node->pos[2];
    np.check_norm = 0;
    np.check_from_diffuse = 0;
    np.max = nphotons;
    np.found = 0;
    np.got_heap = 0;
    np.dist2[0] = max_dist2_pre;

    if (useCompression){
      float norm[3];
      photon_norm(norm, node);
    
      np.norm[0] = norm[0];
      np.norm[1] = norm[1];
      np.norm[2] = norm[2];
    }

    // locate the nearest photons
    locate_photons( &np, 1 );

    // if less than 8 photons return
    float irrad[3] = {0};
    if (np.found<8){
      store_photon_floats( irrad, node->irrad );
      continue;
    }

    float pdir[3];
    float ndir[3];
    float r = sqrt(np.dist2[0]);
    float power[3];

    const float alpha = 0.918*2.0; //0.918;
    const float beta = 1.953;
    float wpg;

    // compute average distance
    avgdist2 += np.dist2[0];
    avgdisti++;
    
    // sum irradiance from all photons
    for (int i=1; i<=np.found; i++) {
      const Photon *p = np.index[i];
      // the photon_dir call and following if can be omitted (for speed)
      // if the scene does not have any thin surfaces

      float dp = sqrt(np.dist2[i]);
      //wpc = 1.0-dp/(k*r);
      wpc = 1.0;


//       // GAUSSIAN IS BROKEN - NORMALIZATION IS NOT CORRECT
//       wpg =
//
// 	alpha * (
//	  
// 	1.0 -
//
// 	( 1.0 -
// 	  exp( -beta * np.dist2[i] / (2.0*np.dist2[0]) )
// 	  )
//	
// 	/
//	
// 	( 1.0 - exp(-beta) )
//	
// 	);


      // 2d adjustment
      if (gv->useDimensions<3){
	photon_norm( ndir, p);
	if (ndir[1]<1.0)
	  // wpg = wpg * 1/cos(angle with plane y=0)
	  wpc /= sqrt(1.0-ndir[1]*ndir[1]);
      }

      photon_dir( pdir, p );
      photon_norm( ndir, node);
      if ( (pdir[0]*ndir[0]+pdir[1]*ndir[1]+pdir[2]*ndir[2]) < 0.0f ) {

	get_photon_floats( power, p->power );
	irrad[0] += power[0]*wpc; // *wpg...
	irrad[1] += power[1]*wpc;
	irrad[2] += power[2]*wpc;
      }
    }

    // const float wpc_norm = (1.-2./(3.*k));
    const float wpc_norm = 1.0;
    const float tmp=(1.0f/M_PI)/( wpc_norm * np.dist2[0]);   // estimate of density
    
    irrad[0] *= tmp;
    irrad[1] *= tmp;
    irrad[2] *= tmp;

    // debugging
    //if (irrad[0]>10.0)
    //  cerr << "Found a bad photon!\n";

    store_photon_floats( irrad, node->irrad );

    // figure purposes
    if (show){// && drand48()<0.3){
	//&& node->pos[0]<99.0 && node->pos[2]<99.0){
      for (int i=0; i<3; i++){
        // irrad[i] *= 1000; //286.0*0.8;
	irrad[i] = 1.0-exp(-2.0*irrad[i]*1000.0);
	irrad[i] = pow(irrad[i], (float)(1.0/2.0));
      }

      if (gv->figureMode){
	
	// set color brightness
	SbColor color(irrad);
	float hue, sat, val;
	color.getHSVValue(hue, sat, val);
	color.setHSVValue(hue, sat, 1.0);
	
	// show as sphere
	showPoint(SbVec3f(node->pos), .1, color);
      }
      else
	showPhoton(SbVec3f(node->pos), SbColor(irrad));
      
    }
  }

  // status message
  if (thread==0)
    fprintf(stderr, "\r%.2f%%\n", 100.0);

  // compute average distance of furthest photon
  if (avgdisti==0)
    cerr << "ERROR: No photons found at " << __FILE__ << ":"
      << __LINE__ << endl;
  else 
    avgdist2 /= avgdisti;
}

// irradiance_estimate_pre returns a precomputed irradiance estimate
// at a given surface position
//
//========================================================
void Photon_map::irradiance_estimate_pre(
  float irrad[3],                // returned irradiance
  const float pos[3],            // surface position
  const float normal[3] ) const  // surface normal at pos
//========================================================
{
  irrad[0] = irrad[1] = irrad[2] = 0.0;
  if (stored_photons==0) return;

  // maximum radius from Frank Suykens PhD dissertation, chapter 8
  // http://www.cs.kuleuven.ac.be/~graphics/CGRG.PUBLICATIONS/
  //   FRANKPHD/suykens_8.pdf
  const float alpha = 0.1;
  const int nphotons = 1;
  
  NearestPhotons np;
  np.dist2 = (float*)alloca( sizeof(float)*(nphotons+1) );
  np.index = (const Photon**)alloca( sizeof(Photon*)*(nphotons+1) );

  np.pos[0] = pos[0];
  np.pos[1] = pos[1];
  np.pos[2] = pos[2];
  np.norm[0] = normal[0];
  np.norm[1] = normal[1];
  np.norm[2] = normal[2];
  np.check_norm = 1;          // must be of similar normal
  np.check_from_diffuse = 0;
  np.max = nphotons;
  np.found = 0;
  np.got_heap = 0;
  np.dist2[0] = max_dist2_pre;

  // locate the nearest photons
  locate_photons( &np, 1 );

  // if less than 1 photon return
  if (np.found==0)
    return;

  const Photon *p = np.index[np.found];
  get_photon_floats( irrad, p->irrad );
  
  return;
}



// get_flux_map returns a map *proporitional to* the incoming flux (one chan).
// Described in Jensen's paper "Importance Driven Path Tracing
// Using the Photon Map".
//
//===============================
void Photon_map::get_flux_map(
  float *flux_map,               // returned irradiance (u, v)
  const int n_theta,             // number of divisions in theta (u)
  const int n_phi,               // number of divisions in phi (v)
  const float pos[3],            // surface position
  const float normal[3],         // surface normal at pos
  const int nphotons ) const     // number of photons to use
//===============================
{
  // reset the map
  for (int u=0; u<n_theta; u++)
    for (int v=0; v<n_phi; v++)
      flux_map[u*n_phi+v] = 0.0;

  if (stored_photons == 0) return;
  
  //float max_dist = max_radius(nphotons);
  float max_dist = 3.0f*max_radius(nphotons);
  
  NearestPhotons np;
  np.dist2 = (float*)alloca( sizeof(float)*(nphotons+1) );
  np.index = (const Photon**)alloca( sizeof(Photon*)*(nphotons+1) );

  np.pos[0] = pos[0]; np.pos[1] = pos[1]; np.pos[2] = pos[2];
  np.check_norm = 0;
  np.max = nphotons;
  np.found = 0;
  np.check_from_diffuse = 1;        // only indirect photons
  np.got_heap = 0;
  np.dist2[0] = max_dist*max_dist;

  // for compressing sphere into disc
  //np.sbr.setValue(SbVec3f(0.0, 1.0, 0.0), SbVec3f(normal));

  // locate the nearest photons
  locate_photons( &np, 1 );

  // if less than 8 photons return
  if (np.found<8)
    return;

  float r = sqrt(np.dist2[0]);
  float pow[3];
  float pdir[3];
  float uf, vf, theta, phi;
  int u, v;
  SbRotation sbr;
  SbVec3f rotateTo = SbVec3f(0.0, 1.0, 0.0);
  SbVec3f fluxDir;
  
  // sum irradiance from all photons
  for (int i=1; i<=np.found; i++) {
    const Photon *p = np.index[i];

    // check that photon is indirect
    if (p->light_id)
      continue;
    
    // the photon_dir call and following if can be omitted (for speed)
    // if the scene does not have any thin surfaces
    photon_dir( pdir, p );
    if ( (pdir[0]*normal[0]+pdir[1]*normal[1]+pdir[2]*normal[2]) < 0.0f ) {

      // rotate pdir to local hemisphere coordinates, and reverse it.
      sbr.setValue(SbVec3f(normal), rotateTo);
      sbr.multVec(-SbVec3f(pdir), fluxDir);
      fluxDir.normalize();

      //showPhoton(SbVec3f(p->pos), SbColor(0.8, 0.8, 1.0));

      // calculate theta
      SbVec3f newFlux;
      theta = acos(fluxDir[1]);

      // calculate phi
      // this is tricky because of acos() return value,
      // floating point error, and divide by zero if theta=0.
      float temp;
      if (theta == 0)
	phi = 0.0;
      else {

	// phi = acos(fluxDir[0]/sin(theta),
	// but sometimes the inner fraction exceeds the valid
	// domain of acos().
	
	temp = fluxDir[0]/sin(theta);

	// detect if this ratio is not in [-1,1]
	if (temp<-1.0)
	  phi = M_PI;
	else if (temp>=1.0)
	  phi = 0.0;
	else {

	  // calculate phi
	  phi = acos(temp);

	  // range of acos() is only [0,pi] instead of [0,2pi],
	  // so we compensate by adjusting if Z direction is negative.
	  if (fluxDir[2]<0.0)
	    phi = 2.0*M_PI-phi;
	}
      }
      
      // calculate u and v
      uf = fluxDir[1];    // cos(theta), but save trig function call
      uf *= uf;
      vf = phi/(2.0*M_PI);

      // convert to int
      if (uf==1.0)              // sometimes uf=1 (theta=0), which is too high
	u = n_theta-1;
      else 
	u = (int)(uf*n_theta);
      v = (int)(vf*n_phi);
      
      // uncompress photon power
      get_photon_floats(pow, p->power);

      // add flux to map (one channel, ignore scale)
      flux_map[u*n_phi+v] += pow[0] + pow[1] + pow[2];
    }
  }

}


void Photon_map::print_stats(){
  float power[3];
  double avg[3] = {0.0, 0.0, 0.0};
  for (int i=1; i<=stored_photons; i++) {  
    get_photon_floats(power, photons[i].power);
    avg[0] += power[0];
    avg[1] += power[1];
    avg[2] += power[2];
  }

  cerr << "***************************************\n";
  cerr << "Statistics for " << stored_photons << " photons (out of "
       << max_photons << " max) :\n";
  cerr << "Total power (" << avg[0] << "," << avg[1] << "," << avg[2] << ")\n";
  avg[0] /= stored_photons;
  avg[1] /= stored_photons;
  avg[2] /= stored_photons;
  cerr << "Average power (" << avg[0] << "," << avg[1] << "," << avg[2] << ")\n";

  // compute variance now that we have mean
  double var[3] = {0.0, 0.0, 0.0};
  for (int i=1; i<=stored_photons; i++) {  
    get_photon_floats(power, photons[i].power);
    var[0] += (power[0]-avg[0])*(power[0]-avg[0]);
    var[1] += (power[1]-avg[1])*(power[1]-avg[1]);
    var[2] += (power[2]-avg[2])*(power[2]-avg[2]);
  }
  var[0] /= stored_photons;
  var[1] /= stored_photons;
  var[2] /= stored_photons;
  cerr << "Variance (" << var[0] << "," << var[1] << "," << var[2] << ")\n";

  // some sample photons
  get_photon_floats(power, photons[1].power);
  cerr << "Photon[1] = " << power[0] << "," << power[1] << "," << power[2] << ")\n";
  get_photon_floats(power, photons[stored_photons/2].power);
  cerr << "Photon[" << stored_photons/2 << "] = "
       << power[0] << "," << power[1] << "," << power[2] << ")\n";
  get_photon_floats(power, photons[stored_photons].power);
  cerr << "Photon[" << stored_photons << "] = "
       << power[0] << "," << power[1] << "," << power[2] << ")\n";
  
}


// photon map power dump
//
// To use:
//
// uncomment line in Pane.cxx to call Photon_map::dump_map();
// run pane. pane dumps photon map in photon.dump.
// run this in matlab:
//
// cd (DIR)
// [hid,mess] = fopen('photon.dump', 'r');
// [h,cnt] = fread(hid, 'single');
// figure; semilogy(sort(h)); fclose(hid);

void Photon_map::dump_map(){
  const char *fn = "photon.dump";
  FILE *file = fopen(fn, "wb");
  if (!file){
    cerr << "Unable to open file " << fn << ": " << strerror(errno) << endl;
    return;
  }

  float power[3];
  double avg[3] = {0.0, 0.0, 0.0};
  for (int i=1; i<=stored_photons; i++) {  
    get_photon_floats(power, photons[i].power);
    float f = MAX3(power[0], power[1], power[2]);
    fwrite(&f, sizeof(float), 1, file);
  }
  fclose(file);

  cerr << "Wrote " << stored_photons << " photons to " << fn << endl;
  return;
}




// irradiance_estimate computes an irradiance estimate
// at a given surface position
//
//======================================
void Photon_map::volume_irradiance_estimate(
  float irrad[3],                // returned irradiance
  const float pos[3],            // surface position
  const int nphotons ) const     // number of photons to use
//======================================
{
  irrad[0] = irrad[1] = irrad[2] = 0.0;
  if (stored_photons == 0) return;
  
  NearestPhotons np;
  np.dist2 = (float*)alloca( sizeof(float)*(nphotons+1) );
  np.index = (const Photon**)alloca( sizeof(Photon*)*(nphotons+1) );

  np.pos[0] = pos[0];
  np.pos[1] = pos[1];
  np.pos[2] = pos[2];
  np.check_norm = 0;
  np.check_from_diffuse = 0;
  np.max = nphotons;
  np.found = 0;
  np.got_heap = 0;
  np.dist2[0] = 100.0;//max_dist2_pre; // FIXME?

  // locate the nearest photons
  locate_photons( &np, 1 );

  // if less than 8 photons return
  if (np.found<8)
    return;

  // photon filtering from jensen p.82
  const float gauss_norm = 1.8790;            // normalization of gaussian
  const float alpha = 0.918 * gauss_norm;
  const float beta = 1.953;
  float wpg;
  
  const float k = 1.1;
//  float wpc;
  float r = sqrt(np.dist2[0]);
  
  float pdir[3];
  float pow[3];

  // temp
  float ndir[3];
  
  // sum irradiance from all photons
  for (int i=1; i<=np.found; i++) {
    const Photon *p = np.index[i];

      // tent filter
//       float dp = sqrt(np.dist2[i]);
//       wpc = 1.0-dp/(k*r);
      
//       // gaussian filter
//       wpg = alpha *
// 	(1.0 -
// 	 ( 1.0 -  exp( -beta * np.dist2[i] / (2.0*np.dist2[0]) ) ) /
// 	 ( 1.0 - exp(-beta) )
// 	);

      get_photon_floats(pow, p->power);

      // 2d adjustment
//       if (gv->useDimensions<3){
	
// 	photon_norm( ndir, p);
// 	if (ndir[1]<1.0)
// 	  wpg /= sqrt(1.0-ndir[1]*ndir[1]);
//       }

      irrad[0] += pow[0];
      irrad[1] += pow[1];
      irrad[2] += pow[2];
  }
  
  //const float wpc_norm = (1.-2./(3.*k));
  //const float wpc_norm = 1.0;

  // 1/(4/3 pi r^3) = (3/4)/pi / r^3
  float tmp = (3.0f/4.0f/M_PI)/(r*r*r);  // estimate of density
    
  irrad[0] *= tmp;
  irrad[1] *= tmp;
  irrad[2] *= tmp;


//   static float minR = 1e30;
//   static float maxI = -1e30;
//   if (r<minR) {
//     minR = r;
//     cerr << "Minr: " << r << " tmp: " << tmp
// 	 << " irrad: " << irrad[R] << " " << irrad[G] << " " << irrad[B]
// 	 << "\n";
//   }
//   if (irrad[0]>maxI){
//     maxI = irrad[0];
//     cerr << "Maxi: " << maxI << "\n";
//   }
    
  
}
