////////////////////////////////////////////////////////////////////////////
//
//WWWW''WW''''''''''WW''''''''''''''''''''''''''''''''''''''''''''WWW'''''''
//WW  W WW          WW                                             WW       
//WW  W WWW   WW   WWWW  WW   WWW       WWWW W   WWW WWW      WWW  WW   WWW 
//WWWW  WW W WW W   WW  WW W  WW W      WW WW W W WW WW W    WW W  WW  W WW 
//WW    WW W WW W   WW  WW W  WW W      WW WW W W WW WW W    WW    WW  W WW 
//WW    WW W  WW    WW   WW   WW W WWWW WW WW W  WWW WWW      WWW WWWW  WWW 
//                                                   WW                     
//'''''''''''''''''''''''''''''''''''''''''''''''''''WW'''''''''''''''''''''
//
// File: Photon_map.h
// Description: Photon_map class header
// Created: 2003/07/17 09:33:17
// Author: Kevin Beason <beason@cs.fsu.edu>
//
// Copyright (c) 2003-2004 Kevin Beason
// 
// Photon map class for the Pane ray tracer
// For more information on Pane, visit:
// http://www.csit.fsu.edu/~beason/pane/
// or see the accompanying documentation
// in the distribution.
//
// class Photon_map implements the photon map
// data structure as described by Henrik Jensen
// in the Photon Mapping algorithm described
// in his book:
//
// "Realistic Image Synthesis
// Using Photon Mapping" by Henrik Wann Jensen 2001
// http://graphics.ucsd.edu/~henrik/papers/book/
//
// Briefly, "photons" (particles representing radiant
// flux) are stored along with their position and
// incoming direction in a kd-tree data structure,
// which, when balanced, provides fast nearest-neighbor
// searches.
//
// Using this nearest-neighbor search, a small
// neighborhood of photons can be found, and using
// the area they subtend an approximate estimate
// of the irradiance can be calculated.
//
// This source code is based on an implementation
// from that book. However, it has been extended a
// fair amount.
// 
// Differences:
//   * irradiance can be precomputed
//   * automatic maximum search radius 
//   * irradiance estimate uses cone filter
//   * photons store normal and precomputed irradiance
//   * photon power and irradiance are compressed, so
//     photon size is actually 2 bytes smaller (26 vs. 28).
//   * "bug" fixed in locate_photons to find exact nearest
// 
// $Id: Photon_map.h 1186 2007-11-12 09:02:06Z beason $
//
////////////////////////////////////////////////////////////////////////////



//------------------------------------------------------------
// Photon_map.h
// An example implementation of the photon map data structure
//
// Henrik Wann Jensen - February 2001
//------------------------------------------------------------

#ifndef PHOTONMAPCLASS
#define PHOTONMAPCLASS

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
#include <pthread.h>
#include "Main.h"
#include "Ray.h"
#include "debug.h"
#include "GlobalVars.h"

using namespace std;

// forward declaration of class Photon_map
class Photon_map;

// for passing object functions to threads
typedef struct pmMemberFunction {
  Photon_map *object;
  void (Photon_map::*pmf)(float &, int, int, int, int);
  int nphotons;
  int thread;
  int num_threads;
  int show;
  float avgdist2;
} pmMemberFunction;


// This is the photon
// The power is not compressed so the
// size was 28 bytes
// size is now 26 bytes
//
//======================
typedef struct Photon {
//======================
  float pos[3];                // photon position
  short plane;                 // splitting plane for kd-tree
  unsigned char theta, phi;    // incoming direction
  unsigned char power[4];      // photon power (compressed)
  unsigned char irrad[4];      // precomputed irradiance (compressed)
  unsigned char normt, normp;  // normal at surface
  char light_id;               // light ID if direct, or 0 for indirect
  char from_diffuse;           // whether photon is from a diffuse reflection
} Photon;


// This structure is used only to locate the
// nearest photons
//
//==============================
typedef struct NearestPhotons {
//==============================
  int max;
  int found;
  int got_heap;
  float pos[3];
  int check_norm;
  int check_from_diffuse;
  float norm[3];
  float *dist2;
  const Photon **index;

  // for compressing sphere into disc
//  SbRotation sbr;
} NearestPhotons;


// This is the Photon_map class
//
//=================
class Photon_map {
//=================
public:

  // constructor
  Photon_map( const int max_phot, GlobalVars *gv_ptr );

  // deconstructor
  ~Photon_map();

  // global variables object
  GlobalVars *gv;

  // stores a photon in the map
  void store(
    const float power[3],        // photon power
    const float pos[3],          // photon position
    const float dir[3],          // photon direction
    const float normal[3],       // surface normal at photon
    const int light_id,          // light ID, or 0 for indirect
    const int from_diffuse,      // from diffuse reflector?
    Photon *buffer = NULL);      // optional buffer to store in instead of map

  // stores a photon in the map, using a buffer for performance
  void buffered_store(
    const float power[3],
    const float pos[3],
    const float dir[3],
    const float normal[3],
    const int light_id,
    const int from_diffuse,
    const int thread = 0,        // thread id
    const int flush = 0);        // flush the buffer?

  // scales photons in map
  void scale_photon_power(
    const float scale );         // 1/(number of emitted photons)

  // balances the kd-tree
  void balance(const int nphotons);  // balance the kd-tree (before use!)

  // computes estimate of the irradiance
  void irradiance_estimate(
    float irrad[3],              // return irradiance
    const float pos[3],          // surface position
    const float normal[3],       // surface normal at pos
    const int nphotons ) const;  // number of photons to use

  // finds nearest photons
  void locate_photons(
    NearestPhotons *const np,    // nm is used to locate the photons
    const int index ) const;     // call with index = 1

  // extracts photon direction
  void photon_dir(
    float *dir,                  // direction of photon (returned)
    const Photon *p ) const;     // the photon

  // extracts photon normal
  void photon_norm(
    float *norm,                 // direction of normal (returned)
    const Photon *p ) const;     // the photon

  // compresses RGB floats into RGBE chars
  void store_photon_floats(      // compresses pow[3] into
    const float *pow,            // photon->(power|irrad)
    unsigned char *p) const;

  // decompresses RGBE chars into RGB floats
  void get_photon_floats(
    float *pow,                  // uncompresses into this pow[3]
    const unsigned char *p) const; // the contents of photon->(power|irrad)

  // gets number of photons in map
  int get_stored_photons(){      // added by kbeason
    return(stored_photons);
  }

  // precompute the irradiance and store at each photon (spawns threads)
  float precompute_irradiance(const int nphotons,
			      const int num_threads=1,
			      const int show=0);

  // precompute irradiance for a subset of photons
  void precompute_irradiance_section(float &avgdist2,
				     const int nphotons,
				     const int thread=0,
				     const int num_threads=1,
				     const int show=0
				     );
  
  // wrapper function for spawning threads
  static void *precompute_wrapper(void *arg){
    pmMemberFunction *memberFunc = static_cast <pmMemberFunction*> (arg);
    Photon_map *pm = memberFunc->object;
    void (Photon_map::*pmf)(float &, int, int, int, int) = memberFunc->pmf;
    (pm->*pmf)(memberFunc->avgdist2,
	       memberFunc->nphotons, memberFunc->thread,
	       memberFunc->num_threads, memberFunc->show
	       );
    return(NULL);
  }

  // gets precompute irradiance
  void irradiance_estimate_pre(
    float irrad[3],                // returned irradiance
    const float pos[3],            // surface position
    const float normal[3] ) const; // surface normal at pos

  void volume_irradiance_estimate(
    float irrad[3],                // returned irradiance
    const float pos[3],            // surface position
    const int nphotons ) const;    // number of photons to use

  // gets histogram map of flux
  void get_flux_map(
    float *flux_map,               // returned irradiance (u, v)
    const int n_theta,             // number of divisions in theta (u)
    const int n_phi,               // number of divisions in phi (v)
    const float pos[3],            // surface position
    const float normal[3],         // surface normal at pos
    const int nphotons ) const;    // number of photons to use

  // maximum radius from Frank Suykens PhD dissertation, chapter 8
  // http://www.cs.kuleuven.ac.be/~graphics/CGRG.PUBLICATIONS/
  //   FRANKPHD/suykens_8.pdf
  float max_radius(const int nphotons) const;

  void compress_sphere(float c=1.0){
    if (c==1.0 || c<=0){
      useCompression = false;
      return;
    }

    cerr << "Enabling volume estimate compression: " << c << "\n";
    sphereCompression = 1./(c*c) - 1;
    useCompression = true;
  }
  
  float bound_rad;   // maximum search radius (bounding sphere of map)
  float bound_rad2;  // maximum search radius, squared
  SbVec3f bbox_cen;
  bool useCompression;       // whether to squish sphere estimate
  float sphereCompression;   // amount to scale irr. est. sphere volume 
                             // in the direction of the normal,
                             // actually c' = (1/c^2-1)

  // debugging functions
  void print_stats();
  void dump_map();       
  
  
private:

  // balances a subtree of the kd-tree
  void balance_segment(
    Photon **pbal,
    Photon **porg,
    const int index,
    const int start,
    const int end );

  // rearranges array subregion to have all photons
  // greater than median in top half, and lower
  // in lower half. a crude sort.
  void median_split(
    Photon **p,
    const int start,
    const int end,
    const int mdian,
    const int axis );

  Photon *photons;           // pointer to photons array

  int stored_photons;        // number of photons stored
  int half_stored_photons;   // precompute half of stored photons
  int max_photons;           // maximum number of photons
  // int prev_scale;            

  float costheta[256];       // cos lookup table for theta
  float sintheta[256];       // sin lookup table for theta
  float cosphi[256];         // cos lookup table for phi
  float sinphi[256];         // sin lookup table for phi

  float bbox_min[3];         // bounding box min
  float bbox_max[3];         // bounding box max

  float max_dist2_pre;       // precomputed maximum search radius for irr_est_pre()

  static const int maxThreads = 16;  // maximum number of threads
  Photon *buffers[maxThreads];       // buffers for storing photons
  int buffPos[maxThreads];           // buffer positions
};

#endif /* PHOTONMAPCLASS */
