#include <cstdlib>
#include <cmath>
#include <iostream>
#include <string.h>
#include <png.h>
#include <getopt.h>
#include "rgbe.h"
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/SbLinear.h>

using namespace std;

// find min/max of three/two vals
#define MIN3(a,b,c) ((((a)<(b))&&((a)<(c))) ? (a) : (((b)<(c)) ? (b) : (c)))
#define MAX3(a,b,c) ((((a)>(b))&&((a)>(c))) ? (a) : (((b)>(c)) ? (b) : (c)))
#define MAX2(a,b) ( ((a)>(b)) ? (a) : (b) )
#define MIN2(a,b) ( ((a)<(b)) ? (a) : (b) )

// clamp definition
// restricts a number to a range [min, max]
template <class T> inline
T clamp(T x, T min, T max){
  if (x<min)
    return(min);
  else if (x>max)
    return(max);
  else
    return(x);
}

#define idx(i,j) (dim[0]*j+i)
#define pix(t,i,j) (texture[t][idx(i,j)])

// component functions
float f0(float t){  return(2.61e6*exp(-(t*t/(.02*.02))));  }
float f1(float t){  float tmp = t+0.02; return(20.91/(tmp*tmp*tmp));  }
float f2(float t){  float tmp = t+0.02; return(72.37/(tmp*tmp));  }
float f3(float t, float l){
  float temp = t-3.0*l/568.0;
  return (436.9 * 568.0/l * exp(-19.75*(temp*temp)));
}

// composite point spread functions

SbColor Pp(float t){
  float a = 0.384*f0(t) + 0.478*f1(t) + 0.138*f2(t);
  return (SbColor(a,a,a));
}
SbColor Pp_lh(float t){
  return(SbColor(0.0, 0.0, 0.0));
}

SbColor Pm(float t){
  float a = 0.368*f0(t) + 0.478*f1(t) + 0.138*f2(t);
  return(SbColor(a,a,a));
}
SbColor Pm_lh(float t){ // lenticular halo
  return(0.016*SbVec3f(f3(t, 611), f3(t, 549), f3(t, 467))); 
}

SbColor Ps(float t){
  float a = 0.282*f0(t) + 0.478*f1(t) + 0.207*f2(t);
  return(SbColor(a,a,a));
}
SbColor Ps_lh(float t){ // lenticular halo
  return(0.033*SbVec3f(f3(t, 611), f3(t, 549), f3(t, 467)));
}

SbColor (*Px)(float t);

int imageHeight = 300;
float imageHeightAngleDeg = 45.0;
int filterSize = 251;
float pixdeg; // = 1.5*1.0/(2.0*M_PI*2.0)*360.0/1200.0; // degrees

SbColor f(float x, float y){
  float dist = sqrt(x*x+y*y);
  float deg = dist*pixdeg;
  return((*Px)(deg));
}

inline float maxc(SbVec3f(x)){
  return (MAX3(x[0], x[1], x[2]));
}
inline float minc(SbVec3f(x)){
  return (MIN3(x[0], x[1], x[2]));
}

int main(int argc, char *argv[]){

  srand48(0);
  int numLines = 400;

  // read options
  static struct option long_options[] = {
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
  };
    
  while (1){
    int c = getopt(argc, argv, "d:f:i:hn:");
    
    if (c == -1)
      break;

    switch (c){
      
    case 'd':
      imageHeightAngleDeg = atof(optarg);
      break;
      
    case 'h':
      cout << "Usage: " << argv[0]
	   << " [-d deg] [-i imageHeightPix] [-f filterSizePix] [-n lines] "
	   << "[photopic|mesopic|scotopic]\n";
      exit(1);
      break;

    case 'i':
      imageHeight = atoi(optarg);
      break;

    case 'f':
      filterSize = atoi(optarg);
      break;
      
    case 'n':
      numLines = atoi(optarg);
      break;

    default:
      cerr << flush;
      fprintf(stderr, "Error: Unknown command line option.\n"
	      "(getopt returned character code 0%o ?" "?)\n", c);
      exit(2);
    }
  }

  // read last option to determine filter type
  enum {PHOTOPIC, MESOPIC, SCOTOPIC};
  char name[3][50] = { "photopic", "mesopic", "scotopic" };
  int filtType = PHOTOPIC;
  if (optind < argc) {
    char *typeStr = argv[optind++];
    if (strcmp(typeStr, "photopic") == 0)
      filtType = PHOTOPIC;
    else if (strcmp(typeStr, "mesopic") == 0)
      filtType = MESOPIC;
    else if (strcmp(typeStr, "scotopic") == 0)
      filtType = SCOTOPIC;
    else {
      cerr << "Unknown filter type: " << typeStr << endl;
      exit(3);
    }
  }


  // draw some lines
  char outFile[50];
  char hdrFile[50];
  sprintf(outFile, "%s.png", name[filtType]);
  sprintf(hdrFile, "%s.hdr", name[filtType]);

  // initialize variables
  pixdeg = imageHeightAngleDeg/(float)imageHeight;
  if (filterSize<1)
    filterSize = 251;
  if (filterSize%2==0)
    filterSize++;
  
  int dim[2];
  dim[0] = dim[1] = filterSize;

  // create two blank pix maps
  SbColor *texture[2];
  texture[0] = new SbColor[dim[0]*dim[1]];
  texture[1] = new SbColor[dim[0]*dim[1]];
  for (int j=0; j<dim[1]; j++)
    for (int i=0; i<dim[0]; i++){
      pix(0,i,j) = SbColor(0.0,0.0,0.0);
      pix(1,i,j) = SbColor(0.0,0.0,0.0);
    }

  cout << "numLines = " << numLines << endl;
  float lineRad = 1.0;

  if (numLines == 0){
    for (int j=0; j<dim[1]; j++)
      for (int i=0; i<dim[0]; i++){
	pix(0,i,j) = SbColor(1.0,1.0,1.0);
	pix(1,i,j) = SbColor(1.0,1.0,1.0);
      }
  }
  else {
     
    // draw two sets of lines
    cout << "Creating " << name[filtType] << " filter.\n";
    for (int t=0; t<2; t++){
      for (int i=0; i<numLines; i++){

	float lineIntens = drand48();
      
	float theta;
	//theta = drand48()*2.0*M_PI;
	theta = (i+drand48())/numLines*2.0*M_PI;  // stratified

	int dj = (theta<=M_PI) ? 1 : -1;
	int j = (int)(dim[1]/2.0 - (lineRad-0.5)*dj);

	while (j>=0 && j<dim[1]){
	  float jminf = (j+0.0-dim[1]/2.0);
	  float jmaxf = (j+1.0-dim[1]/2.0);
	  float iminf = jminf/tan(theta);
	  float imaxf = jmaxf/tan(theta);

	  // swap if iminf>imaxf
	  if (iminf>imaxf){
	    float temp = iminf;
	    iminf = imaxf;
	    imaxf = temp;
	  }
    
	  float dif = fabs(lineRad/sin(theta));
	  int imin = (int)(iminf-dif+0.5+dim[0]/2.0);
	  int imax = (int)(imaxf+dif-0.5+dim[0]/2.0);
	  //cerr << imax-imin << endl;
    
	  // loop over pixels in row
	  for (int i=MAX2(imin, 0); i<MIN2(imax+1, dim[0]); i++){

	    // find pixel center in spatial coordinates
	    SbVec2f p;
	    p[0] = i-dim[0]/2.0+0.5;
	    p[1] = j-dim[1]/2.0+0.5;

	    // find distance of p to line
	    SbVec2f d = SbVec2f(cos(theta), sin(theta));
	    SbVec2f o = SbVec2f(0.0, 0.0);

	    float dp = (p-o).dot(d);
	    float dist;
	    if (dp>0.0)
	      dist = ((p-o) - dp*d).length();
	    else 
	      dist = (p-o).length();
	
	    // tent filter
	    if (dist>=lineRad)
	      continue;
	    float filt = 1.0/lineRad - 1.0/(lineRad*lineRad)*dist;
	    float f = MAX2((p-o).length()*2*M_PI/numLines, 1.0/numLines); // 
	    float val = filt*f*lineIntens;
	    pix(t,i,j) += SbColor(val,val,val);

	    // max value allowed is f
	    //if (pix(t,i,j)[0]>f)
	    //  pix(t,i,j).setValue(f,f,f);
	  }

	  // next row
	  j += dj;
	}
      
      }

      
    // NO NEED - scalar change is removed when normalized later
/*    
    // normalize lines
    // c*Sum(f)/n = Avg(f) = 1
    // c = n/Sum(f)
    // f* = c*f = n/Sum(f) * f
    SbColor sum(0.0, 0.0, 0.0);
    for (int j=0; j<dim[1]; j++)
      for (int i=0; i<dim[0]; i++)
        sum += pix(t,i,j);

    SbColor c(1.0, 1.0, 1.0);
    int n = dim[1]*dim[2];
    for (int i=0; i<3; i++)
      if (sum[i]>0.0)
	c[i] = n/sum[i];

    for (int j=0; j<dim[1]; j++)
      for (int i=0; i<dim[0]; i++){
	pix(t,i,j)[0] *= c[0];
	pix(t,i,j)[1] *= c[1];
	pix(t,i,j)[2] *= c[2];
      }
*/    
    
    }
  }
  
  // multiply by point spread function
  // evaluate two functions: P, and P_lh
  cerr << "\n";
  for (int t=0; t<2; t++){

    // set up filter function pointers
    switch (filtType){
    case PHOTOPIC:
      if (t==0)
	Px = Pp;
      else
	Px = Pp_lh;
      break;
    case MESOPIC:
      if (t==0)
	Px = Pm;
      else
	Px = Pm_lh;
      break;
    case SCOTOPIC:
      if (t==0)
	Px = Ps;
      else
	Px = Ps_lh;
      break;
    }
    
    for (int j=0; j<dim[1]; j++){

      // status line
      cerr << "\r" << j/(float)(dim[1]-1)*100.0 << "%    ";
    
      for (int i=0; i<dim[0]; i++){

	// perform 2D trapezoidal integration
	float a = (float)i-dim[0]/2.0, b=a+1.0;
	float c = (float)j-dim[0]/2.0, d=c+1.0;
      
	// determine four corners, find ratio
	float max, min, tmp_max, tmp_min;
	SbColor tmp, sum1;

	// sum four corners and keep track of min and max
	tmp = f(a,c);
	sum1 = tmp;
	max = maxc(sum1);
	min = minc(sum1);
	
	tmp = f(b,c);
	sum1+=tmp;
	tmp_max = maxc(tmp); if (tmp_max>max) max = tmp_max;
	tmp_min = minc(tmp); if (tmp_min<min) min = tmp_min;
	
	tmp = f(a,d);
	sum1+=tmp;
	tmp_max = maxc(tmp); if (tmp_max>max) max = tmp_max;
	tmp_min = minc(tmp); if (tmp_min<min) min = tmp_min;
	
	tmp = f(b,d);
	sum1+=tmp;
	tmp_max = maxc(tmp); if (tmp_max>max) max = tmp_max;
	tmp_min = minc(tmp); if (tmp_min<min) min = tmp_min;

	// choose number of samples based on ratio of min and max
	int samps = 3; // (max-min)/min > 2.0 ? 100 : 10;  // FIXME
	int n = samps-1;

	float h = (b-a)/n;
	float k = (d-c)/n;

	SbVec3f sum2 = SbVec3f(0.0, 0.0, 0.0);
	for (int k=1; k<n; k++){
	  float x_k = (b-a)/(float)n*k+a;
	  float y_k = (d-c)/(float)n*k+c;
	  sum2 += f(x_k, c) + f(x_k, d) + f(a, y_k) + f(b, y_k);
	}
	sum1 += 2.0*sum2;

	sum2 = SbVec3f(0.0, 0.0, 0.0);
	for (int k=1; k<n; k++)
	  for (int l=1; l<n; l++){
	    float x_k = (b-a)/(float)n*k+a;
	    float y_l = (d-c)/(float)n*l+c;
	    sum2 += f(x_k, y_l);
	  }
	sum1 += 4.0*sum2;

	sum1 *= 0.25*h*k;

	pix(t,i,j)[0] *= sum1[0];
	pix(t,i,j)[1] *= sum1[1];
	pix(t,i,j)[2] *= sum1[2];
      }
    }
    cerr << "\n";
  }

  // multiply and normalize
  SbColor sum = SbColor(0.0, 0.0, 0.0);
  for (int j=0; j<dim[1]; j++)
    for (int i=0; i<dim[0]; i++){
      pix(0,i,j) += pix(1,i,j);
      sum += pix(0,i,j);
    }

  SbVec3f mul = SbVec3f(1.0, 1.0, 1.0);
  if (sum[0]>0.0)
    mul[0] = 1.0/sum[0];
  if (sum[1]>0.0)
    mul[1] = 1.0/sum[1];
  if (sum[2]>0.0)
    mul[2] = 1.0/sum[2];
  cout << "Sum = " << sum[0] << ", " << sum[1] << ", " << sum[2] << "\n";
  cout << "Mul = ("
       << mul[0] << ","
       << mul[1] << ","
       << mul[2] << ")"
       << endl;

  float gmin = 1e10, gmax = -1e10;
  float sum2 = 0.0;
  for (int j=0; j<dim[1]; j++)
    for (int i=0; i<dim[0]; i++) {
      pix(0,i,j)[0] *= mul[0];
      pix(0,i,j)[1] *= mul[1];
      pix(0,i,j)[2] *= mul[2];
      
      // debugging
      float sum1 = pix(0,i,j)[0];
      sum2 += sum1;
      if (sum1>0.0){
	sum1 = log10(sum1);
	if (sum1<gmin) gmin = sum1;
	if (sum1>gmax) gmax = sum1;
      }
    }


  // debugging
  cerr << "min, max of log10() = " << gmin << ", " << gmax << "\n";
  cerr << " sum=" << sum2 << "\n";

  // write image

  // write hdr
  FILE *f = fopen(hdrFile, "wb");
  RGBE_WriteHeader(f, dim[0], dim[1], NULL);
  for (int j=dim[1]-1; j>=0; j--)
    for (int i=0; i<dim[0]; i++){
      RGBE_WritePixels(f, texture[0][j*dim[0]+i].getValue(), 1);
      
      // write binary float gray map
      //fwrite(texture[0][j*dim[0]+i].getValue(), sizeof(float), 1, f);
    }
  fclose(f);

  // adjust (brighten) for png
  //gmin = -10; gmax = -3.3;
  for (int j=0; j<dim[1]; j++)
    for (int i=0; i<dim[0]; i++) {
      SbColor c = pix(0,i,j);
      for (int k=0; k<3; k++){
	if (c[k]>0.0){
	  c[k] = log10(c[k]);
	  //c[k] = (c[k]-gmin)/(gmax-gmin+0.05);
	  if ((gmax-gmin)>0.0)
	    c[k] = (c[k]-gmin)/(gmax-gmin);
	}
      }
      pix(0,i,j) = c;
    }

  // write png

  // open file
  FILE *fp = fopen(outFile, "wb");
  if (!fp){
    cerr << "Error writing file " << outFile << endl;
    return(1);
  }

  // create png_struct and png_info
  png_structp png_ptr = png_create_write_struct
    (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!png_ptr){
    cerr << "Error creating not create png_ptr." << endl;
    return(1);
  }

  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (!info_ptr){
    png_destroy_write_struct(&png_ptr,
			     (png_infopp)NULL);
    cerr << "Error creating PNG data structures." << endl;
    return(1);
  }

  // set up error handling. PNG will expect to longjmp() here
  if (setjmp(png_jmpbuf(png_ptr))){
       png_destroy_write_struct(&png_ptr, &info_ptr);
       fclose(fp);
    cerr << "PNG creation error." << endl;
    return(1);
  }

  // set up png output
  png_init_io(png_ptr, fp);
  png_set_IHDR(png_ptr, info_ptr, dim[0], dim[1],
               8, PNG_COLOR_TYPE_RGB,
	       PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT,
	       PNG_FILTER_TYPE_DEFAULT);

  /*
  // convert comment to c style string
  int comment_size = comment.size();
  char commentstr[comment_size+1];
  strcpy(commentstr, comment.c_str());
  
  // create comment
  png_text text_ptr[1];
  text_ptr[0].key = "Comment";
  text_ptr[0].text = commentstr;
  text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
  png_set_text(png_ptr, info_ptr, text_ptr, 1);

  // set rendering intent
  //png_set_sRGB(png_ptr, info_ptr, PNG_sRGB_INTENT_ABSOLUTE);
  
  // set gamma
  png_set_gAMA(png_ptr, info_ptr, 1.0/gv->gamma);

  // set chromacity info (values from srgb.com)
  png_set_cHRM(png_ptr, info_ptr,
	       .3127, .3290,        // white point x, y
	       .64, .33,            // r 
	       .15, .06,            // g
	       .3127, .3290);       // b

  */

  // create array of bytes
  png_byte *image = new png_byte[dim[0]*dim[1]*3];
  png_bytep row_pointers[dim[1]];
  SbColor *c;
  float r, g, b;
  for (int j=0; j<dim[1]; j++){
    for (int i=0; i<dim[0]; i++){
      c = &texture[0][dim[0]*j+i];

      unsigned char rc, gc, bc;
      rc = (unsigned char) clamp((int)((*c)[0]*255.0), 0, 255);
      gc = (unsigned char) clamp((int)((*c)[1]*255.0), 0, 255);
      bc = (unsigned char) clamp((int)((*c)[2]*255.0), 0, 255);

      image[j*(dim[0]*3)+i*3+0] = rc;
      image[j*(dim[0]*3)+i*3+1] = gc;
      image[j*(dim[0]*3)+i*3+2] = bc;
    }
  }
  for (int k=0; k<dim[1]; k++)
    row_pointers[k] = &image[(dim[1]-k-1)*dim[0]*3];

  // write image
  png_write_info(png_ptr, info_ptr);
  png_write_image(png_ptr, row_pointers);
  png_write_end(png_ptr, info_ptr);

  // clean up
  png_destroy_write_struct(&png_ptr, &info_ptr);
  return(0);
}
